#include <linux/console.h>
#include <linux/types.h>
#include <linux/wait.h>

#include "speakup.h"
#include "spk_priv.h"

#define SYNTH_BUF_SIZE 8192	/* currently 8K bytes */

static u_char synth_buffer[SYNTH_BUF_SIZE];	/* guess what this is for! */
static u_char *buff_in = synth_buffer;
static u_char *out_start = synth_buffer;
static u_char *buff_out = synth_buffer;
static u_char *buffer_end = synth_buffer + SYNTH_BUF_SIZE - 1;

/* These try to throttle applications by stopping the TTYs
 * Note: we need to make sure that we will restart them eventually, which is
 * usually not possible to do from the notifiers. TODO: it should be possible
 * starting from linux 2.6.26.
 *
 * So we only stop when we know alive == 1 (else we discard the data anyway),
 * and the alive synth will eventually call start_ttys from the thread context.
 */
void speakup_start_ttys(void)
{
	int i;

	for (i = 0; i < MAX_NR_CONSOLES; i++) {
		if (speakup_console[i] && speakup_console[i]->tty_stopped)
			continue;
		if ((vc_cons[i].d) && (vc_cons[i].d->port.tty)) {
			start_tty(vc_cons[i].d->port.tty);
			speakup_console[i]->tty_stopped = 0;
		}
	}
}
EXPORT_SYMBOL_GPL(speakup_start_ttys);

static void speakup_stop_ttys(void)
{
	int i;

	for (i = 0; i < MAX_NR_CONSOLES; i++) {
		if (speakup_console[i] && speakup_console[i]->tty_stopped == 0)
			continue;
		if ((vc_cons[i].d) && (vc_cons[i].d->port.tty)) {
			stop_tty(vc_cons[i].d->port.tty);
			speakup_console[i]->tty_stopped = 1;
		}
	}
}

int synth_buffer_status(bool clear_flush)
{
	int bufsize;
	if (speakup_info.flushing) {
		bufsize = -1;
		if (clear_flush)
			speakup_info.flushing = 0;
	} else if (buff_in >= buff_out)
		bufsize = buff_in - buff_out;
	else
		bufsize = SYNTH_BUF_SIZE - (buff_out - buff_in);
	out_start = buff_out;
	// check status after get means we are done writing the segment
	return bufsize;
}
EXPORT_SYMBOL_GPL(synth_buffer_status);

void synth_buffer_add(char ch)
{
	int bytes_free;
	if (!synth->alive)
		return; // don't stop ttys if no synth
	if (buff_in >= out_start)
		bytes_free = SYNTH_BUF_SIZE - (buff_in - out_start);
	else
		bytes_free = out_start - buff_in;
	if (bytes_free <= 100) {
		synth_start();
		speakup_stop_ttys();
	}
	if (bytes_free <= 1)
		return;
	*buff_in++ = ch;
	if (buff_in > buffer_end)
		buff_in = synth_buffer;
}

void synth_write(const char *buf, int count)
{
	int bytes_free;
	if (!synth->alive)
		return; // don't stop ttys if no synth
	if (buff_in >= out_start)
		bytes_free = SYNTH_BUF_SIZE - (buff_in - out_start);
	else
		bytes_free = out_start - buff_in;
	if (bytes_free <= 100) {
		synth_start();
		speakup_stop_ttys();
	}
	if (count < 0)
		count = strlen(buf);
	if (bytes_free <= count)
		return;
	while (count--) {
		*buff_in++ = *buf++;
		if (buff_in > buffer_end)
			buff_in = synth_buffer;
	}
	synth_start();
}

int synth_buffer_gets(char *buf, int maxchars)
{
	int count = 0;

	while (count < maxchars &&buff_out != buff_in) {
		buf[count++] = *buff_out++;
		if (buff_out > buffer_end)
			buff_out = synth_buffer;
	}
	return count;
}
EXPORT_SYMBOL_GPL(synth_buffer_gets);


int synth_buffer_ungets(int num_chars)
{
	int ctr;
	for (ctr = 0; ctr < num_chars; ctr++) {
		if (buff_out == out_start)
			break;
		buff_out--;
		if (buff_out < synth_buffer)
			buff_out = buffer_end;
	}
	return ctr;
}
EXPORT_SYMBOL_GPL(synth_buffer_ungets);

char synth_buffer_getc(void)
{
	char ch;

	if (buff_out == buff_in)
		return 0;
	ch = *buff_out++;
	if (buff_out > buffer_end)
		buff_out = synth_buffer;
	return ch;
}
EXPORT_SYMBOL_GPL(synth_buffer_getc);

void synth_buffer_ungetc(void)
{
	if (buff_out == out_start)
		return;
	buff_out--;
	if (buff_out < synth_buffer)
		buff_out = buffer_end;
}
EXPORT_SYMBOL_GPL(synth_buffer_ungetc);

char synth_buffer_peek(void)
{
	return (buff_out == buff_in) ? 0 : *buff_out;
}
EXPORT_SYMBOL_GPL(synth_buffer_peek);

void synth_buffer_clear(void)
{
	buff_in = buff_out;
}
EXPORT_SYMBOL_GPL(synth_buffer_clear);
