/* speakup.c
 * review functions for the speakup screen review package.
 * originally written by: Kirk Reiser and Andy Berdan.
 *
 * extensively modified by David Borowski.
 *
 ** Copyright (C) 1998  Kirk Reiser.
 *  Copyright (C) 2003  David Borowski.
 *  Copyright (C) 2016  David Borowski.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <linux/kernel.h>
#include <linux/vt.h>
#include <linux/tty.h>
#include <linux/mm.h>		/* __get_free_page() and friends */
#include <linux/vt_kern.h>
#include <linux/ctype.h>
#include <linux/selection.h>
#include <linux/unistd.h>
#include <linux/jiffies.h>
#include <linux/kthread.h>
#include <linux/keyboard.h>	/* for KT_SHIFT */
#include <linux/kbd_kern.h>	/* for vc_kbd_* and friends */
#include <linux/input.h>
#include <linux/kmod.h>

/* speakup_*_selection */
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/consolemap.h>

#include <linux/spinlock.h>
#include <linux/notifier.h>

#include <linux/uaccess.h>	/* copy_from|to|user() and others */

#include "spk_priv.h"
#include "speakup.h"

typedef int (*done_hand) (struct vc_data *d, u_short val);
typedef void (*spkup_hand) (struct vc_data *d);
#define MAX_DELAY msecs_to_jiffies(500)
#define MINECHOCHAR 33

MODULE_AUTHOR("Kirk Reiser <kirk@braille.uwo.ca>");
MODULE_AUTHOR("Daniel Drake <dsd@gentoo.org>");
MODULE_AUTHOR("David Borowski <david.a.borowski@gmail.com>");
MODULE_DESCRIPTION("Speakup console speech");
MODULE_LICENSE("GPL");
MODULE_VERSION(SPEAKUP_VERSION);

char *synth_name;
module_param_named(synth, synth_name, charp, S_IRUGO);
module_param_named(quiet, spk_quiet_boot, bool, S_IRUGO);

MODULE_PARM_DESC(synth, "Synth to start if speakup is built in.");
MODULE_PARM_DESC(quiet, "Do not announce when the synthesizer is found.");

special_func spk_special_handler;

short spk_pitch_shift, synth_flags;
#define MAXLINE 254
#define LINEBUFFERS 3
static char line_buffer[LINEBUFFERS][MAXLINE+2];
static u_char attrib_buffer[MAXLINE+2];
static char *buf = line_buffer[0];
static int line_buffer_len[LINEBUFFERS], currbuf = 0, lastbuf = LINEBUFFERS-1;
static u_short read_all_mode = 0, read_all_key = 0;
static u_short key_speakup = 0, spk_key_locked = 0, update_ctr;
static int cursor_updated, update_start, update_end, color_count;
int spk_attrib_bleep, spk_bleeps;
int spk_cursor_time, spk_bleep_time = 10;
int spk_no_intr, spk_spell_delay;
int spk_key_echo, spk_say_word_ctl;
int spk_say_ctrl, spk_bell_pos;
int spk_punc_mask, spk_reading_mask;
char spk_str_caps_start[MAXVARLEN+1] = "\0";
char spk_str_caps_stop[MAXVARLEN+1] = "\0";
const struct st_bits_data spk_punc_info[] = {
	{"none", "", 0},
	{"some", "/$%&@", SOME},
	{"most", "$%&#()=+*/@^<>|\\", MOST},
	{"all", "!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~", PUNC},
	// none and all are not editable.
	{"delimiters", "", B_WDLM},
	{"repeats", "()", CH_RPT},
	{"extended numeric", "", B_EXNUM},
	{"symbols", "", B_SYM},
	{NULL, NULL}
};

static char mark_cut_flag;
static u_char spk_shift_table[SHIFT_TBL_SIZE];
const u_char spk_key_defaults[] =
#include "speakupmap.h"
u_char spk_key_buf[KEY_BUFF_SIZE];
#define MAX_KEY 160
u_char *spk_our_keys[MAX_KEY];

/* Speakup Cursor Track Variables */

/* cursor track modes, must be ordered same as cursor_msgs */
enum {
	CT_Off = 0, CT_On, CT_Smart,
	CT_Highlight, CT_Window, DO_MULTI, DO_ROUTE
};
#define CT_Max DO_MULTI

enum { // read all modes
SENTENCE_MODE = 1, LINE_MODE, READ_MULTI, PAUSED = 8
};

void spkup_write(const char *in_buf, int count, u_short mask);
static int
stop_readall(struct vc_data *vc, u_char type, u_char ch, u_short key);

static char *phonetic[] = {
	"alfa", "bravo", "charlie", "delta", "echo", "foxtrot",
	"golf", "hotel", "india", "juliett", "keelo", "leema",
	"mike", "november", "oscar", "papa", "keh beck", "romeo",
	"sierra", "tango", "uniform", "victer", "whiskey", "x ray",
	"yankee", "zulu"
};

/* array of 256 char pointers (one for each character description)
 * initialized to default_chars and user selectable via
 * /proc/speakup/characters
 */
char *spk_characters[256];

char *spk_default_chars[256] = {
/*000*/ "null", "^eigh", "^b", "^c", "^d", "^e", "^f", "^g",
/*008*/ "^h", "^i", "^j", "^k", "^l", "^m", "^n", "^o",
/*016*/ "^p", "^q", "^r", "^s", "^t", "^u", "^v", "^w",
/*024*/ "^x", "^y", "^zed", "control", "control", "control", "control",
	    "control",
/*032*/ "space", "bang!", "quote", "number", "dollar", "percent", "and",
	    "tick",
/*040*/ "left paren", "right paren", "star", "plus", "comma", "dash",
	    "dot",
	"slash",
/*048*/ "zero", "one", "two", "three", "four", "five", "six", "seven",
	"eight", "nine",
/*058*/ "colon", "semmy", "less", "equals", "greater", "question", "at",
/*065*/ "EIGH", "B", "C", "D", "E", "F", "G",
/*072*/ "H", "I", "J", "K", "L", "M", "N", "O",
/*080*/ "P", "Q", "R", "S", "T", "U", "V", "W", "X",
/*089*/ "Y", "ZED", "left bracket", "backslash", "right bracket",
	    "caret",
	"line",
/*096*/ "accent", "eigh", "b", "c", "d", "e", "f", "g",
/*104*/ "h", "i", "j", "k", "l", "m", "n", "o",
/*112*/ "p", "q", "r", "s", "t", "u", "v", "w",
/*120*/ "x", "y", "zed", "left brace", "bar", "right brace", "tihlduh",
/*127*/ "del", "control", "control", "control", "control", "control",
	    "control", "control", "control", "control", "control",
/*138*/ "control", "control", "control", "control", "control",
	    "control", "control", "control", "control", "control",
	    "control", "control",
/*150*/ "control", "control", "control", "control", "control",
	    "control", "control", "control", "control", "control",
/*160*/ "nbsp", "inverted bang",
/*162*/ "cents", "pounds", "currency", "yen", "broken bar", "section",
/*168*/ "diaeresis", "copyright", "female ordinal", "double left angle",
/*172*/ "not", "soft hyphen", "registered", "macron",
/*176*/ "degrees", "plus or minus", "super two", "super three",
/*180*/ "acute accent", "micro", "pilcrow", "middle dot",
/*184*/ "cedilla", "super one", "male ordinal", "double right angle",
/*188*/ "one quarter", "one half", "three quarters",
	    "inverted question",
/*192*/ "EIGH GRAVE", "EIGH ACUTE", "EIGH CIRCUMFLEX", "EIGH TILDE",
			"EIGH OOMLAUT", "EIGH RING",
/*198*/ "AE", "C CIDELLA", "E GRAVE", "E ACUTE", "E CIRCUMFLEX",
	    "E OOMLAUT",
/*204*/ "I GRAVE", "I ACUTE", "I CIRCUMFLEX", "I OOMLAUT", "ETH",
	    "N TILDE",
/*210*/ "O GRAVE", "O ACUTE", "O CIRCUMFLEX", "O TILDE", "O OOMLAUT",
/*215*/ "multiplied by", "O STROKE", "U GRAVE", "U ACUTE",
	    "U CIRCUMFLEX",
/*220*/ "U OOMLAUT", "Y ACUTE", "THORN", "sharp s", "a grave",
/*225*/ "a acute", "a circumflex", "a tilde", "a oomlaut", "a ring",
/*230*/ "ae", "c cidella", "e grave", "e acute",
/*234*/ "e circumflex", "e oomlaut", "i grave", "i acute",
	    "i circumflex",
/*239*/ "i oomlaut", "eth", "n tilde", "o grave", "o acute",
	    "o circumflex",
/*245*/ "o tilde", "o oomlaut", "divided by", "o stroke", "u grave",
	    "u acute",
/* 251 */ "u circumflex", "u oomlaut", "y acute", "thorn", "y oomlaut"
};

/* array of 256 u_short (one for each character)
 * initialized to default_chartab and user selectable via
 * /sys/module/speakup/parameters/chartab
 */
u_short spk_chartab[256];
EXPORT_SYMBOL_GPL(spk_chartab);

static u_short default_chartab[256] = {
	B_CTL+B_WDLM, B_CTL, B_CTL, B_CTL, B_CTL, B_CTL, B_CTL, B_CTL,	/* 0-7 */
	B_CTL, B_CTL, A_CTL, B_CTL, B_CTL, B_CTL, B_CTL, B_CTL,	/* 8-15 */
	B_CTL, B_CTL, B_CTL, B_CTL, B_CTL, B_CTL, B_CTL, B_CTL,	/*16-23 */
	B_CTL, B_CTL, B_CTL, B_CTL, B_CTL, B_CTL, B_CTL, B_CTL,	/* 24-31 */
	WDLM, A_PUNC, PUNC, PUNC, PUNC, PUNC, PUNC, A_PUNC,	/*  !"#$%&' */
	PUNC, PUNC, PUNC, PUNC, A_PUNC, A_PUNC, A_PUNC, PUNC,	/* ()*+, -./ */
	NUM, NUM, NUM, NUM, NUM, NUM, NUM, NUM,	/* 01234567 */
	NUM, NUM, A_PUNC, A_PUNC, PUNC, PUNC, PUNC, A_PUNC,	/* 89:;<=>? */
	PUNC, A_CAP, A_CAP, A_CAP, A_CAP, A_CAP, A_CAP, A_CAP,	/* @ABCDEFG */
	A_CAP, A_CAP, A_CAP, A_CAP, A_CAP, A_CAP, A_CAP, A_CAP,	/* HIJKLMNO */
	A_CAP, A_CAP, A_CAP, A_CAP, A_CAP, A_CAP, A_CAP, A_CAP,	/* PQRSTUVW */
	A_CAP, A_CAP, A_CAP, PUNC, PUNC, PUNC, PUNC, PUNC,	/* XYZ[\]^_ */
	PUNC, ALPHA, ALPHA, ALPHA, ALPHA, ALPHA, ALPHA, ALPHA,	/* `abcdefg */
	ALPHA, ALPHA, ALPHA, ALPHA, ALPHA, ALPHA, ALPHA, ALPHA,	/* hijklmno */
	ALPHA, ALPHA, ALPHA, ALPHA, ALPHA, ALPHA, ALPHA, ALPHA,	/* pqrstuvw */
	ALPHA, ALPHA, ALPHA, PUNC, PUNC, PUNC, PUNC, 0,	/* xyz{|}~ */
	B_CAPSYM, B_CAPSYM, B_SYM, B_SYM, B_SYM, B_SYM, B_SYM, /* 128-134 */
	B_SYM,	/* 135 */
	B_SYM, B_SYM, B_SYM, B_SYM, B_SYM, B_SYM, B_SYM, /* 136-142 */
	B_CAPSYM,	/* 143 */
	B_CAPSYM, B_CAPSYM, B_SYM, B_CAPSYM, B_SYM, B_SYM, B_SYM, /* 144-150 */
	B_SYM,	/* 151 */
	B_SYM, B_SYM, B_CAPSYM, B_CAPSYM, B_SYM, B_SYM, B_SYM, /*152-158 */
	B_SYM,	/* 159 */
	WDLM, B_SYM, B_SYM, B_SYM, B_SYM, B_SYM, B_CAPSYM, /* 160-166 */
	B_SYM,	/* 167 */
	B_SYM, B_SYM, B_SYM, B_SYM, B_SYM, B_SYM, B_SYM, B_SYM,	/* 168-175 */
	B_SYM, B_SYM, B_SYM, B_SYM, B_SYM, B_SYM, B_SYM, B_SYM,	/* 176-183 */
	B_SYM, B_SYM, B_SYM, B_SYM, B_SYM, B_SYM, B_SYM, B_SYM,	/* 184-191 */
	A_CAP, A_CAP, A_CAP, A_CAP, A_CAP, A_CAP, A_CAP, A_CAP,	/* 192-199 */
	A_CAP, A_CAP, A_CAP, A_CAP, A_CAP, A_CAP, A_CAP, A_CAP,	/* 200-207 */
	A_CAP, A_CAP, A_CAP, A_CAP, A_CAP, A_CAP, A_CAP, B_SYM,	/* 208-215 */
	A_CAP, A_CAP, A_CAP, A_CAP, A_CAP, A_CAP, A_CAP, ALPHA,	/* 216-223 */
	ALPHA, ALPHA, ALPHA, ALPHA, ALPHA, ALPHA, ALPHA, ALPHA,	/* 224-231 */
	ALPHA, ALPHA, ALPHA, ALPHA, ALPHA, ALPHA, ALPHA, ALPHA,	/* 232-239 */
	ALPHA, ALPHA, ALPHA, ALPHA, ALPHA, ALPHA, ALPHA, B_SYM,	/* 240-247 */
	ALPHA, ALPHA, ALPHA, ALPHA, ALPHA, ALPHA, ALPHA, ALPHA	/* 248-255 */
};

struct task_struct *speakup_task;
static u_char spk_lastkey, speakup_key;
u_char spk_close_press;
EXPORT_SYMBOL_GPL(spk_close_press);

struct st_spk_t *speakup_console[MAX_NR_CONSOLES], *current_spk;

DEFINE_MUTEX(spk_mutex);

static int keyboard_notifier_call(struct notifier_block *,
				  unsigned long code, void *param);

static struct notifier_block keyboard_notifier_block = {
	.notifier_call = keyboard_notifier_call,
};

static int vt_notifier_call(struct notifier_block *,
			    unsigned long code, void *param);

static struct notifier_block vt_notifier_block = {
	.notifier_call = vt_notifier_call,
};

static unsigned char get_attributes(struct vc_data *vc, int pos)
{
	u16 ch = scr_readw(screen_pos(vc, pos, 1));
	return ch>>8;
}


static void speakup_date(struct vc_data *vc)
{
	spk_x = vc->vc_x;
	spk_y = vc->vc_y;
	spk_pos = (vc->vc_pos-vc->vc_origin)/2;
	spk_old_attr = spk_attr;
	spk_attr = get_attributes(vc, spk_pos);
}

static short bleeps[4], bleep_index = 0;
static const short bleep_freqs[] = {
	350, 370, 392, 414, 440, 466, 491, 523, 554, 587, 619, 659
};

static void do_bleep(u_long data);
static DEFINE_TIMER(bleep_timer, do_bleep, 0, 0);

static void do_bleep(u_long data)
{
	short val, bleep_freq;
	if (bleep_index <= 0)
		return;
	val = bleeps[--bleep_index];
	if (val >= 0) {
		bleep_freq = bleep_freqs[val % 12];
		if (val > 11)
			bleep_freq *= (1 << (val / 12));
		kd_mksound(bleep_freq, msecs_to_jiffies(spk_bleep_time));
	}
	if (bleep_index > 0)
		mod_timer(&bleep_timer, jiffies+msecs_to_jiffies(spk_bleep_time+20));
}


static void bleep(short val)
{
	bleep_index = 1;
	bleeps[0] = val;
	mod_timer(&bleep_timer, jiffies+msecs_to_jiffies(10));
}

static void bleep3(short val1, short val2, short val3)
{
	del_timer(&bleep_timer);
	bleep_index = 0;
	if (val3 > -1) // -1 is 2 bleeps
		bleeps[bleep_index++] = val3;
	bleeps[bleep_index++] = val2; // allow -1 for pause between bleeps
	bleeps[bleep_index++] = val1;
	mod_timer(&bleep_timer, jiffies+msecs_to_jiffies(10));
}

static void speakup_shut_up(struct vc_data *vc)
{
	spk_shut_up |= 0x01;
	spk_parked &= 0xf0;
	speakup_date(vc);
	spk_do_flush();
}

static void speech_kill(struct vc_data *vc)
{
	int do_init = (spk_killed) ? 1 : 0;
	char val = synth->is_alive(synth, do_init);

	if (val == 0)
		return;

	/* re-enables synth, if disabled */
	if (val == 2 || spk_killed) {
		spk_shut_up &= ~0x40;
		spk_msg_say(MSG_IAM_ALIVE);
	} else {
		spk_msg_say(MSG_YOU_KILLED_SPEAKUP);
		spk_shut_up |= 0x40;
	}
}

static void speakup_off(struct vc_data *vc)
{
	if (spk_shut_up & 0x80) {
		spk_shut_up &= 0x7f;
		spk_msg_say(MSG_HEY_THATS_BETTER);
	} else {
		spk_shut_up |= 0x80;
		spk_msg_say(MSG_YOU_TURNED_ME_OFF);
	}
	speakup_date(vc);
}

static void speakup_parked(struct vc_data *vc)
{
	if (spk_parked & 0x80) {
		spk_parked = 0;
		spk_msg_say(MSG_UNPARKED);
	} else {
		spk_parked |= 0x80;
		spk_msg_say(MSG_PARKED);
	}
}

static void speakup_cut(struct vc_data *vc)
{
	static const char err_buf[] = "set selection failed";
	int ret, buf_num = 0;
	static int ps, pe;
  static struct vc_data *mark_cons = NULL;

	pe = spk_pos;
	if (!mark_cut_flag) {
		mark_cut_flag = 1;
		ps = pe;
		mark_cons = vc;
		spk_msg_say(MSG_MARK);
		return;
	}
	if (mark_cons != vc) {
		synth_write_string("selection, mark console not the same as cut");
		return;
	}
	mark_cut_flag = 0;
	buf_num = speakup_key-CUT;
	spk_msg_say2(FUNC_ID, speakup_key);
	if (buf_num == 0)
		buf_num = fg_console+3;

	ret = speakup_set_selection(vc, ps, pe, buf_num);
	switch (ret) {
	case 0:
		break;
	case -EFAULT :
		pr_warn("%sEFAULT\n", err_buf);
		break;
	case -EINVAL :
		pr_warn("%sEINVAL\n", err_buf);
		break;
	case -ENOMEM :
		pr_warn("%sENOMEM\n", err_buf);
		break;
	}
}

static void speakup_paste(struct vc_data *vc)
{
	struct tty_struct *tty = vc->port.tty;
	int buf_num = 0;
	if (mark_cut_flag) {
		mark_cut_flag = 0;
		spk_msg_say(MSG_MARK_CLEARED);
		return;
	}
	buf_num = speakup_key-PASTE;
	spk_msg_say2(FUNC_ID, speakup_key);
	if (buf_num == 0)
		buf_num = fg_console+3;
	speakup_paste_selection(tty, buf_num);
}

static void read_paste(struct vc_data *vc)
{
	int saved_punc_mask = spk_punc_mask;
	int buf_num = speakup_key-PASTE_READ;
	if (buf_num == 0)
		buf_num = fg_console+3;
	spk_punc_mask = spk_reading_mask;
	speakup_read_selection(buf_num);
	spk_punc_mask = saved_punc_mask;
}

static void say_attributes(struct vc_data *vc)
{
	int fg = spk_attr & 0x0f;
	int bg = spk_attr >> 4;

	if (fg > 8) {
		spk_msg_say2(COLOR_ID, BRIGHT);
		fg &= 7;
	}
	spk_msg_say2(COLOR_ID, fg);
	if (bg > 8) {
		spk_msg_say2(COLOR_ID, BLINKING);
		bg &= 7;
	} else
		spk_msg_say_sp(MSG_ON);
	spk_msg_say2(COLOR_ID, bg);
}

enum {
	edge_notset, edge_quiet,
	edge_top, edge_bottom,
	edge_left, edge_right
};

static void announce_edge(int msg_id)
{
	if (spk_bleeps & 1)
		bleep(spk_y);
	if ((spk_bleeps & 2) && (msg_id >= edge_top))
		spk_msg_say(MSG_EDGE_TOP + msg_id-edge_top);
}

static void speak_char(u_char ch)
{
	char *cp = spk_characters[ch];
	struct var_t *direct = spk_get_var(DIRECT);

	if (direct && direct->u.n.value) {
		if (IS_CHAR(ch, B_CAP)) {
			spk_pitch_shift++;
			synth_printf( "%s%c%s", spk_str_caps_start, ch, spk_str_caps_stop);
		} else
			synth_buffer_add(ch);
		return;
	}
	if (cp == NULL) {
		pr_info("speak_char: cp == NULL!\n");
		return;
	}
	if (IS_CHAR(ch, B_CAP)) {
		spk_pitch_shift++;
		synth_printf(" %s%s%s ", spk_str_caps_start, cp, spk_str_caps_stop);
	} else if (*cp == '^') {
		synth_printf(" %s %s ", spk_msg_get(MSG_CTRL), ++cp);
	} else {
		synth_printf(" %s ", cp);
	}
}

static u16 get_char(struct vc_data *vc, int pos, u_char *attribs)
{
	u16 ch, w = ' ', c;
	w = scr_readw(screen_pos(vc, pos, 1));
c = w & 0xff;
if (w & vc->vc_hi_font_mask)
	c |= 0x100;
	ch = inverse_translate(vc, c, 0);
	if (attribs)
		*attribs = (w & 0xff00) >> 8;
	return ch;
}

static void say_char(struct vc_data *vc)
{
	u_short ch;

	spk_old_attr = spk_attr;
	ch = get_char(vc, spk_pos, &spk_attr);
	if (spk_attr != spk_old_attr) {
		if (spk_attrib_bleep & 1)
			bleep3((spk_attr&0x0f), (spk_attr&0xf0)>>4, -1);
		if (spk_attrib_bleep & 2)
			say_attributes(vc);
	}
	speak_char(ch & 0xff);
}

static void say_phonetic_char(struct vc_data *vc)
{
	u_short ch;

	spk_old_attr = spk_attr;
	ch = get_char(vc, spk_pos, &spk_attr);
	if (isascii(ch) && isalpha(ch)) {
		ch &= 0x1f;
		synth_printf("%s\n", phonetic[--ch]);
	} else {
		if (IS_CHAR(ch, B_NUM))
			spk_msg_say_sp(MSG_NUMBER);
		speak_char(ch);
	}
}

static void say_prev_char(struct vc_data *vc)
{
	if (spk_x == 0) {
		announce_edge(edge_left);
		return;
	}
	spk_pos--;
	say_char(vc);
}

static void say_next_char(struct vc_data *vc)
{
	if (spk_x == vc->vc_cols-1) {
		announce_edge(edge_right);
		return;
	}
	spk_pos++;
	say_char(vc);
}

static void say_char_above(struct vc_data *vc)
{
	if (spk_y == 0) {
		announce_edge(edge_top);
		return;
	}
	spk_pos -= vc->vc_cols;
	spk_y--;
	say_char(vc);
}

static void say_char_below(struct vc_data *vc)
{
	if (spk_y == vc->vc_rows-1) {
		announce_edge(edge_bottom);
		return;
	}
	spk_pos += vc->vc_cols;
	spk_y++;
	say_char(vc);
}

// 0 = not delimiter, 1 = word delim bit set in chartab and 2 = space/null
static int is_delimiter( char ch )
{
	if ( ch == SPACE || ch == 0 )
		return 2;
	return ((spk_chartab[(u_char)ch]&B_WDLM)) ? 1 : 0;
}

/* get_word - will first check to see if the character under the
 * reading cursor is a space and if spk_say_word_ctl = 2 it will
 * return the word space.  If spk_say_word_ctl is < 2 it will check to
 * see if there is a word starting on the next position to the right
 * and return that word if it exists.  If it does not exist it will
 * move left to the beginning of any previous word on the line or the
 * beginning of the line whichever comes first..
 */

static int get_word(struct vc_data *vc, int start, int end)
{
	int tmp_pos = spk_pos, cnt = 0;
	char ch, ch1 = SPACE;
	u_char attr, attr1;
	int state = 0, last_state;

	spk_old_attr = spk_attr;
	ch = (char)get_char(vc, tmp_pos, &attr);

	if (spk_say_word_ctl > 1 && ch == SPACE) {
		*buf = '\0';
		spk_msg_say(MSG_SPACE);
		return 0;
	}
	state = is_delimiter(ch);
	if (tmp_pos < end-1) {
		ch1 = get_char(vc, tmp_pos+1, &attr1);
		if (state && !is_delimiter(ch1))
			goto get_w1;
		if (spk_say_word_ctl > 0 && attr != attr1)
			goto get_w1;
	}
	while (--tmp_pos >= start) {
		last_state = state;
		attr1 = attr;
		state = is_delimiter(get_char(vc, tmp_pos, &attr));
		if (last_state > 0) continue;
		if (state > 0 || (spk_say_word_ctl > 0 && attr != attr1))
			break;
	}
get_w1:
	tmp_pos++; // at beginning of word
	if (spk_say_word_ctl > 1 && tmp_pos > start && state == 1) {
		int dlm_pos = tmp_pos;
		// scan back over delimiters and if we hit space, say the delims also
// otherwise they are said after the previous word
		while (--dlm_pos >= start)  {
			state = is_delimiter(get_char(vc, dlm_pos, &attr));
			if (state != 1) break;
		}
		if (state == 2) {
			while (++dlm_pos < tmp_pos)
				buf[cnt++] = (char)get_char(vc, dlm_pos, &spk_attr);
		}
	}
	ch = (char)get_char(vc, tmp_pos, &spk_attr);
	state = is_delimiter(ch);
	buf[cnt++] = ch;
	while (++tmp_pos <= end) {
		last_state = state;
		ch = (char)get_char(vc, tmp_pos, &attr);
		state = is_delimiter(ch);
		if (state == 2 || (spk_say_word_ctl > 0 && attr != attr1))
			break;
		if (last_state > 0 && state == 0)
			break;
		buf[cnt++] = ch;
		if (state > 0 && spk_say_word_ctl < 2)
			break; // only one delimiter if not say_word_ctl
	}
	line_buffer_len[currbuf] = cnt;
	buf[cnt] = '\0';
	return cnt;
}

static void say_word_at(struct vc_data *vc, int start, int end)
{
	int cnt = get_word(vc, start, end);
	if (cnt == 0)
		return;
	buf[cnt++] = SPACE;
	spkup_write(buf, cnt, spk_reading_mask);
}

static void say_word(struct vc_data *vc)
{
	int start = spk_pos-spk_x;
	int end = start+vc->vc_cols-1;
	say_word_at(vc, start, end);
}

static void say_prev_word(struct vc_data *vc)
{
	int start, end;
	u_short edge_said = edge_notset, last_state = 2, state = 2;
	u_char attr, attr1;

	if (spk_pos == 0) {
		announce_edge(edge_top);
		return;
	}
	if (spk_x == 0) {
		start = spk_pos-vc->vc_cols;
		spk_y--;
	} else
	start = spk_pos-spk_x;
	while (--spk_pos >= 0) {
		last_state = state;
		attr1 = attr;
		state = is_delimiter( get_char(vc, spk_pos, &attr));
		if (spk_pos == start) {
			edge_said = edge_left;
			if (state < 2) {
				last_state = state;
				spk_pos--;
				break;
			}
			start -= vc->vc_cols;
			spk_y--;
		}
		if (last_state > 0) continue;
		if (state > 0 || (spk_say_word_ctl > 0 && attr != attr1))
			break;
	} // we exit loop with spk_pos one char before the word
end = start+vc->vc_cols-1;
	if (++spk_pos == 0)
		edge_said = edge_top;
	if (edge_said > edge_quiet)
		announce_edge(edge_said);
	if (last_state < 2)
		say_word_at(vc, start, end);
}

static void say_next_word(struct vc_data *vc)
{
	int start, end, screen_end = (vc->vc_screenbuf_size/2)-1;
	u_char attr, attr1;
	u_short edge_said = 0, last_state, state;

	if (spk_pos >= screen_end) {
		announce_edge(edge_bottom);
		return;
	}
	start = spk_pos-spk_x;
	end = start+vc->vc_cols-1;
	if (spk_pos >= end) {
		end += vc->vc_cols;
		spk_y++;
	}
		
	state = is_delimiter( get_char(vc, spk_pos, &attr));
	while (++spk_pos <= screen_end) {
		last_state = state;
		attr1 = attr;
		state = is_delimiter( get_char(vc, spk_pos, &attr));
		if (spk_pos == end) {
			edge_said = edge_right;
			end += vc->vc_cols;
			spk_y++;
			if (state == 0) {
				spk_pos++;
				break;
			}
		}
		if (state > 0) continue;
		if (last_state > 0 || (spk_say_word_ctl > 0 && attr != attr1))
			break; // either delimiter to word or attributes changed.
	}
	if (spk_pos > screen_end) {
		spk_pos = screen_end;
		end = screen_end;
		spk_y = vc->vc_rows-1;
		edge_said = edge_bottom;
	}
	start = end-(vc->vc_cols-1);
	if (edge_said > 0)
		announce_edge(edge_said);
	say_word_at(vc, start, end);
}

static void say_word_above(struct vc_data *vc)
{
	if (spk_y == 0) {
		announce_edge(edge_top);
		return;
	}
	spk_pos -= vc->vc_cols;
	spk_y--;
	if (is_delimiter( get_char(vc, spk_pos, NULL)))
		say_prev_word(vc);
	else say_word(vc);
}

static void say_word_below(struct vc_data *vc)
{
	if (spk_y == vc->vc_rows-1) {
		announce_edge(edge_bottom);
		return;
	}
	spk_pos += vc->vc_cols;
	spk_y++;
	if (is_delimiter( get_char(vc, spk_pos, NULL)))
		say_next_word(vc);
	else say_word(vc);
}

static void find_word_end(struct vc_data *vc)
{
	int start, end, screen_end = (vc->vc_screenbuf_size/2)-1;
	u_char attr, attr1;
	u_short edge_said = 0, last_state, state = 2;

	if (spk_pos >= screen_end) {
		announce_edge(edge_bottom);
		return;
	}
	start = spk_pos-spk_x;
	end = start+vc->vc_cols-1;
	if (spk_pos >= end)
		end += vc->vc_cols;
	while (++spk_pos <= screen_end) {
		last_state = state;
		attr1 = attr;
		state = is_delimiter( get_char(vc, spk_pos, &attr));
		if (spk_pos == end) {
			edge_said = edge_right;
			if (state == 0)
				break;
			end += vc->vc_cols;
		}
		if (last_state > 0) continue;
		if (state > 0 || (spk_say_word_ctl > 0 && attr != attr1)) {
			spk_pos--;
			break;
		}
	}
	if (spk_pos > screen_end) {
		spk_pos = screen_end;
		edge_said = edge_bottom;
	}
	if (edge_said > 0)
		announce_edge(edge_said);
	say_char(vc);
}

static void spell_word(struct vc_data *vc)
{
	static char const *delay_str[] = { "", ",", ".", ". .", ". . ." };
	int start = spk_pos-spk_x;
	int end = start+vc->vc_cols-1;
	char *cp = buf, *str_cap = spk_str_caps_stop;
	char *cp1, *last_cap = spk_str_caps_stop;
	u_char ch;

	if (!get_word(vc, start, end))
		return;
	while ((ch = (u_char) *cp)) {
		if (cp != buf)
			synth_printf(" %s ", delay_str[spk_spell_delay]);
		if (IS_CHAR(ch, B_CAP)) {
			str_cap = spk_str_caps_start;
			if (*spk_str_caps_stop)
				spk_pitch_shift++;
			else	/* synth has no pitch */
				last_cap = spk_str_caps_stop;
		} else
			str_cap = spk_str_caps_stop;
		if (str_cap != last_cap) {
			synth_write_string(str_cap);
			last_cap = str_cap;
		}
		if (speakup_key == SPELL_PHONETIC
		    && (isascii(ch) && isalpha(ch))) {
			ch &= 31;
			cp1 = phonetic[--ch];
		} else {
			cp1 = spk_characters[ch];
			if (*cp1 == '^') {
				spk_msg_say_sp(MSG_CTRL);
				cp1++;
			}
		}
		synth_write_string(cp1);
		cp++;
	}
	if (str_cap != spk_str_caps_stop)
		synth_write_string(spk_str_caps_stop);
}

static int get_line(struct vc_data *vc, int line, int from, int to)
{
	int start = line*vc->vc_cols;
	int end = start+to;
	u_char tmp, ch;
	char *cp;
	u_char *ap = attrib_buffer;
	start += from;
	while (--end >= start) {
		ch = (char) get_char(vc, end, &tmp);
		if ( ch != SPACE) break;
	}
	if ( end < start)
		return 0;
	lastbuf = currbuf++;
	if (currbuf >= LINEBUFFERS)
		currbuf = 0;
	buf = line_buffer[currbuf];
	cp = buf;
	spk_old_attr = spk_attr;

	if (end-start >= MAXLINE)
end = start+MAXLINE;
	while (start <= end) {
		*cp++ = (char) get_char(vc, start++, ap);
		ap++;
	}
	spk_attr = *attrib_buffer;
	*cp = '\0';
		line_buffer_len[currbuf] = cp-buf;
	return line_buffer_len[currbuf];
}

// mode bits are: 8 say blank on empty, 4 compare with prev,
// 2 if indent is on say it, 1 use reading punc setting

static int say_from_to(struct vc_data *vc, int line, int from,
						int to, int mode)
{
	int len, mask, cl = current_spk->column_mode, c_len = -1;
	char *cp;
	if (!(mode&16) || line < cl)
	cl = -1;
	if (cl >= 0 && line > cl) {
		if ((c_len = get_line(vc, cl, from, to)) < spk_x)
			cl = -1;
	}
	len = get_line(vc, line, from, to);
	if (len <= 0) {
		if ((mode&8))
			spk_msg_say(MSG_BLANK);
		return 0;
	}
	if ((mode&4)) {
		int n = line_buffer_len[lastbuf];
		if (len == n && !memcmp(buf, line_buffer[lastbuf], len))
			return -1;
	}
	cp = buf;
	mask = ((mode&1)) ? spk_reading_mask : spk_punc_mask;
	if (cl >= 0 && line >= cl && len > c_len-5) {
		char *header;
		int start = spk_x, end = spk_x, pos = -1;
		if (cl == line)
			lastbuf = currbuf; // we are on the header!
		header = line_buffer[lastbuf];
		// look left and right from spk_x for first non-space for closest column
		while (start >= 0 && end < len) {
			if (--start >= 0 && header[start] != SPACE) {
				pos = start;
				break;
			} else if (++end < len && header[end] != SPACE) {
				pos = end;
				break;
			}
		}
		start = pos;
		while (start > 0 && header[start] != SPACE) start--;
		start++;
		pos = start;
		end = pos;
		while (end <len && header[end] != SPACE) end++;
		if (pos != spk_x)
			spk_pos += (pos-spk_x);
		if (currbuf != lastbuf) {
			while (start >= 0 && cp[start] != SPACE) start--;
			start++;
			// and content to the right of header
			while (end <len && cp[end] != SPACE) end++;
		}
		cp += start;
		len = end-start;
	} else if ((mode&2) && (current_spk->indent)) {
		while (*cp == SPACE)
			cp++;
		if (cp > buf)
			synth_printf("%d, ", (int)(cp-buf));
			len -= (cp-buf);
	}
	cp[len++] = '\n';
	spkup_write(cp, len, mask);
	return len;
}

static int say_line_from_to(struct vc_data *vc, int from, int to,
						int mode)
{
	return say_from_to(vc, spk_y, from, to, mode);
}

static void say_line(struct vc_data *vc)
{
	int mode = 27;
	if (speakup_key > SAY_NEXT_LINE)
		mode = 11;
	say_from_to(vc, spk_y, 0, vc->vc_cols, mode);
	// columns if on with no shift, say blank if empty, indent and reading_punc
}

static void say_prev_line(struct vc_data *vc)
{
	if (spk_y == 0) {
		announce_edge(edge_top);
		return;
	}
	spk_pos -= vc->vc_cols;
	spk_y--;
	say_line(vc);
}

static void say_next_line(struct vc_data *vc)
{
	if (spk_y == vc->vc_rows-1) {
		announce_edge(edge_bottom);
		return;
	}
	spk_pos += vc->vc_cols;
	spk_y++;
	say_line(vc);
}

/* Sentence Reading Commands */

#define MARKER_SIZE 32
struct marker {
	u_short pos;
	u_char buf_num, index_num;
};
static struct marker markers[MARKER_SIZE];
static int last_in = 0, last_out = 0;

static void set_index(int buf_num, int pos)
{
	markers[last_out].index_num = synth_insert_index(0);
	markers[last_out].buf_num = buf_num;
	markers[last_out++].pos = (u_short)pos;
	if (last_out >=MARKER_SIZE)
		last_out = 0;
}

static int say_sentence_buf(int buf_num, int startpos)
{
	int n, len = line_buffer_len[buf_num];
	char *start = line_buffer[buf_num], *last_space;
	char *buffend = start+len, *cp;
	u_short char_type = 0;
	n = len/10; // min distance between synth indexes
	// for 80 columns, they can't be closer than 8 chars.
	*buffend++ = (LINE_MODE == read_all_mode) ? '\n' : ' ';
	cp = start+ startpos;
	last_space = cp;
	set_index(buf_num, cp-start);
	while (cp < buffend) {
		u_char ch = (u_char)*cp++;
		if ( ch != SPACE) {
			char_type = spk_chartab[ch];
			continue;
		}
		if ((cp-last_space > n && char_type == A_PUNC) || cp-last_space > 2*n) {
			set_index(buf_num, cp-start);
			spkup_write(last_space, cp-last_space, spk_punc_mask);
			last_space = cp;
		}
		char_type = spk_chartab[ch];
	}
	spkup_write(last_space, cp-last_space, spk_punc_mask);
	return len;
}


static void start_read_all_timer(struct vc_data *vc, u_short cmd, bool send_key);

static done_hand done_handler = NULL;

	static int route_x, route_y, route_pos;

static int do_route(struct vc_data *vc, u_short val)
{
	static int last_move;
	int cursor_move, x_diff = spk_x-route_x, y_diff = spk_y-route_y;
	u_short cmd, sound = -1;
	if (route_pos < spk_pos)
		cursor_move = spk_pos-route_pos;
	else cursor_move = route_pos-spk_pos;
	if (y_diff > 0) {
		cmd = KEY_UP;
		if ((y_diff&3) == 0)
			sound = y_diff;
	} else if (y_diff < 0) {
		cmd = KEY_DOWN;
		if ((y_diff&3) == 0)
			sound = -y_diff;
	} else if (x_diff > 0) {
		cmd = KEY_LEFT;
		if ((x_diff&3) == 0)
			sound = (x_diff/4);
	} else if (x_diff < 0) {
		cmd = KEY_RIGHT;
		if ((x_diff&3) == 0)
			sound = -(x_diff/4);
	} else {
		say_word(vc);
		return 1;
	}
	if (val != READ_ALL && cursor_move >= last_move) {
		spk_msg_say(MSG_ROUTE_FAILED);
		bleep3(0, 12, 24);
		return 1; // distance must get smaller with each cursor move
	}
	last_move = cursor_move;
	if (sound >= 0)
		bleep(sound);
	start_read_all_timer(vc, cmd, true);
	return 2;
}

// for multi line reads don't hold the lock while dumping whole screen.
// this allows other things to schedule including the voice update thread
static int multi_start, multi_end, multi_left, multi_right, multi_mode;

static int continue_multi(struct vc_data *vc, u_short val)
{
	say_from_to(vc, multi_start++, multi_left, multi_right, multi_mode);
	if (multi_start >= multi_end)
		return 1;
	start_read_all_timer(vc, READ_MULTI, false);
	return 2;
}

// additional mode bit 32 means use win_left and win_right as from/to
static void
multi_read(struct vc_data *vc, int start, int end, int mode)
{
	if ((mode&32)) {
		multi_left = win_left;
		multi_right = win_right;
		multi_mode = mode&31;
	} else {
		multi_left = 0;
		multi_right = vc->vc_cols;
		multi_mode = mode;
}
	say_from_to(vc, start++, multi_left, multi_right, multi_mode);
	if (start >= end)
		return;
	multi_start = start;
	multi_end = end;
	done_handler = continue_multi;
	start_read_all_timer(vc, READ_MULTI, false);
}

static void say_screen(struct vc_data *vc)
{
	multi_read(vc, 0, vc->vc_rows, 17);
}

static void say_window(struct vc_data *vc)
{
	if (win_start < 2)
		spk_msg_say(MSG_NO_WINDOW);
	else
		multi_read (vc, win_top, win_bottom+1, 33);
}

static void top_edge(struct vc_data *vc)
{
	spk_pos = spk_x;
	spk_y = 0;
	say_line(vc);
}

static void bottom_edge(struct vc_data *vc)
{
	spk_y = vc->vc_rows-1;
	spk_pos = (spk_y*vc->vc_cols)+spk_x;
	say_line(vc);
}

static void left_edge(struct vc_data *vc)
{
	spk_pos -= spk_x;
	spk_x = 0;
	say_char(vc);
}

static void right_edge(struct vc_data *vc)
{
	spk_pos += (vc->vc_cols - spk_x-1);
	spk_x = vc->vc_cols-1;
	say_char(vc);
}

static void say_first_char(struct vc_data *vc)
{
	int i, len = get_line(vc, spk_y, 0, vc->vc_cols);
	u_char ch;

	if (len == 0) {
		spk_msg_say(MSG_BLANK);
		return;
	}
	for (i = 0; i < len; i++)
		if (buf[i] != SPACE)
			break;
	ch = buf[i];
	spk_pos -= (spk_x - i);
	synth_printf("%d, ", ++i);
	speak_char(ch);
}

static void say_last_char(struct vc_data *vc)
{
	int len = get_line(vc, spk_y, 0, vc->vc_cols);
	u_char ch;

	if (len == 0) {
		spk_msg_say(MSG_BLANK);
		return;
	}
	ch = buf[--len];
	spk_pos -= (spk_x - len);
	synth_printf("%d, ", ++len);
	speak_char(ch);
}

static void say_position(struct vc_data *vc)
{
	if (SAY_LINE_COLUMN == speakup_key)
		synth_printf("%d, %d", spk_y+1, spk_x+1 );
	else
		synth_printf(spk_msg_get(MSG_POS_INFO), spk_y+1, spk_x+1,
			vc->vc_num+1);
	synth_buffer_add('\n');
}

/* Added by brianb */
static void say_char_num(struct vc_data *vc)
{
	u_short ch = get_char(vc, spk_pos, NULL);

	ch &= 0xff;
	synth_printf(spk_msg_get(MSG_CHAR_INFO), ch, ch);
}

static void say_from_top(struct vc_data *vc)
{
	multi_read(vc, 0, spk_y, 17);
}

static void say_to_bottom(struct vc_data *vc)
{
	multi_read(vc, spk_y, vc->vc_rows, 17);
}

static void say_from_left(struct vc_data *vc)
{
	say_line_from_to(vc, 0, spk_x, 9);
}

static void say_to_right(struct vc_data *vc)
{
	say_line_from_to(vc, spk_x, vc->vc_cols, 9);
}


void spkup_write(const char *in_buf, int count, u_short mask)
{
	static int rep_count = 0;
	static u_char ch = '\0', old_ch = '\0';
	static u_short char_type = 0, last_type = 0;
	int in_count = count;
	while (count--) {
		old_ch = ch;
		last_type = char_type;
		ch = (u_char)*in_buf++;
		char_type = spk_chartab[ch];
		if (ch == old_ch && !(char_type&B_NUM)) {
			if (++rep_count > 2)
				continue;
		} else {
			if ((last_type&CH_RPT) && rep_count > 2) {
				synth_buffer_add(' ');
				synth_printf(spk_msg_get(MSG_REPEAT_DESC), ++rep_count);
				synth_buffer_add(' ');
			}
			rep_count = 0;
		}
		if (ch == spk_lastkey) {
			rep_count = 0;
			if (spk_key_echo == 1 && ch >= MINECHOCHAR)
				speak_char(ch);
		} else if (char_type & B_ALPHA) {
			if ((synth_flags & SF_DEC) && (last_type & PUNC) && old_ch != 39)
				synth_buffer_add(SPACE);
			synth_buffer_add(ch);
		} else if (char_type & B_NUM) {
			rep_count = 0;
			synth_buffer_add(ch);
		} else if (char_type&mask) {
			speak_char(ch);
			char_type &= ~PUNC; /* for dec nospell processing */
		} else if (char_type&SYNTH_OK) {
/* these are usually puncts like . and , which synth needs for expression.
 * suppress multiple to get rid of long pausesand clear repeat count so if
 *someone has repeats on you don't get nothing repeated count */
			if (ch != old_ch)
				synth_buffer_add(ch);
			else
				rep_count = 0;
		} else {
			if (old_ch != ch)
				synth_buffer_add(SPACE);
			else
				rep_count = 0;
		}
	}
	spk_lastkey = 0;
	if (in_count > 2 && rep_count > 2) {
		if (last_type&CH_RPT) {
			synth_buffer_add(' ');
			synth_printf(spk_msg_get(MSG_REPEAT_DESC2), ++rep_count);
			synth_buffer_add(' ');
		}
		rep_count = 0;
	}
	synth_start();
}


static void cursor_done(u_long data);
static DEFINE_TIMER(cursor_timer, cursor_done, 0, 0);

static void
console_switch_vars(int new_cons)
{
	struct st_var_header *p_header;
	int i;
	for (i = FIRST_SET_VAR; i < MAXVARS; i++) {
		p_header = spk_get_var_header( i );
		if ( p_header && p_header->var_type == VAR_MULTI )
			spk_set_num_var(new_cons, p_header, E_NEWCON);
	}
	spk_shut_up &= 0xfe;
}

static void
do_handle_shift(struct keyboard_notifier_param *param)
{
	unsigned long flags;
	int vc_num = param->vc->vc_num;
	if ( spk_special_handler)
		return;
	spin_lock_irqsave(&speakup_info.spinlock, flags);
	current_spk = speakup_console[vc_num];
	spk_shut_up &= 0xfe;
	spk_do_flush();
	if (spk_say_ctrl)
		spk_msg_say2(CONTROL_ID, KVAL(param->value));
	spin_unlock_irqrestore(&speakup_info.spinlock, flags);
}

int spk_keybuff_size(void)
{
	return KEY_BUFF_SIZE;
}

int spk_set_key_info(const u_char *key_info)
{
	int i = 0, states, key_data_len, ret = 0;
	const u_char *cp = key_info;
	u_char *cp1 = spk_key_buf;
	u_char ch = 0, version, num_keys;
	if (NULL == cp)
		cp = spk_key_defaults;
	version = *cp++;
	if (version != KEY_MAP_VER)
		return -1;
	num_keys = (int)*cp;
	states = (int)cp[1];
	key_data_len = (states+1) * (num_keys+1);
	if ( key_data_len < 8) return -2;
	key_data_len += 4;
	if (key_data_len >= KEY_BUFF_SIZE)
		return -3;
	memset(spk_shift_table, 0, SHIFT_TBL_SIZE);
	memset(spk_our_keys, 0, sizeof(spk_our_keys));
	cp1 = spk_key_buf;
	memcpy(cp1, cp, key_data_len);
	/* get num_keys, states and data*/
	cp1 += 2; /* now pointing at shift states */
	for (i = 1; i <= states; i++) {
		ch = *cp1++;
		if (ch >= SHIFT_TBL_SIZE)
			return -4;
		if (spk_shift_table[ch] == 0)
			spk_shift_table[ch] = i;
	}
	if (ch == 0) { /* added state for speakup with ctl/shift/alt */
		for (i = 1; i <=31; i++) { // add a catch-all for speakup-key
			if (spk_shift_table[i] == 0)
				spk_shift_table[i] = states;
		}
	}
	cp1++;
	for (i = 0; i  < num_keys; i++) {
	ch = *cp1;
		if (!ch || ch >= MAX_KEY) {
			pr_warn("set key failed: key %d\n", ret);
			return -5;
		}
		spk_our_keys[ch] = cp1;
		cp1 += states+1;
	}
	return 0;
}

static void toggle_cursoring(struct vc_data *vc)
{
	if (++cursor_track >= CT_Max)
		cursor_track = 0;
	spk_msg_say2(FUNC_ID, CURSOR_OFF+ cursor_track);
}

void spk_set_char(int index, char *desc)
{
			if (spk_characters[index] != spk_default_chars[index])
				kfree(spk_characters[index]);
			spk_characters[index] = desc;
}

void spk_reset_default_chars(void)
{
	int i;

	/* First, free any non-default */
	for (i = 0; i < 256; i++) {
		if ((spk_characters[i])
		    && (spk_characters[i] != spk_default_chars[i]))
			kfree(spk_characters[i]);
	}
	memcpy(spk_characters, spk_default_chars, sizeof(spk_default_chars));
}

void spk_reset_default_chartab(void)
{
	memcpy(spk_chartab, default_chartab, sizeof(default_chartab));
}

static const struct st_bits_data *pb_edit = NULL;

static int do_edit_bits(struct vc_data *vc, u_char type, u_char ch, u_short key)
{
	short mask = pb_edit->mask, ch_type = spk_chartab[ch];

	if (type == KT_SHIFT)
		return 0;
	if (key == KEY_SPACE) {
		spk_msg_say(MSG_EDIT_DONE);
		spk_special_handler = NULL;
		return 1;
	}
	if (type != KT_LATIN || (ch_type & B_NUM) || ch < SPACE)
		return -1;
	if (mask < PUNC && !(ch_type & PUNC))
		return -1;
	spk_chartab[ch] ^= mask;
	speak_char(ch);
	spk_msg_say((spk_chartab[ch]&mask) ? MSG_ON : MSG_OFF);
	return 1;
}

/* Allocation concurrency is protected by the console semaphore */
static int speakup_allocate(struct vc_data *vc)
{
	int vc_num = vc->vc_num;
	current_spk = speakup_console[vc_num];
	if (current_spk == NULL) {
		current_spk = kzalloc(sizeof(*speakup_console[0]), GFP_ATOMIC);
		if (current_spk == NULL)
			return -ENOMEM;
		speakup_console[vc_num] = current_spk;
		cursor_track = 1;
		current_spk->column_mode = -1;
		current_spk->w_header = -1;
		speakup_date(vc);
	} else if (!spk_parked)
		speakup_date(vc);
	return 0;
}

static void speakup_deallocate(struct vc_data *vc)
{
	int vc_num;

	vc_num = vc->vc_num;
	if (speakup_console[vc_num]) {
		kfree(speakup_console[vc_num]);
		speakup_console[vc_num] = NULL;
	}
}

static int old_cursor_pos, old_cursor_x, old_cursor_y;
static int cursor_con;


enum {
	RA_NOTHING, RA_TIMER, RA_NOFAKE
};

static void start_read_all_timer(struct vc_data *vc, u_short cmd, bool send_key)
{
	unsigned long interval = spk_cursor_time;
	cursor_updated = 0;
	color_count = 0;
	update_start = 0x7fff;
	update_end = 0;
	cursor_con = vc->vc_num;
	old_cursor_pos = (vc->vc_pos-vc->vc_origin)/2;
	old_cursor_x = vc->vc_x;
	old_cursor_y = vc->vc_y;
	if (KEY_BACKSPACE == cmd || READ_ALL == cmd)
		interval /= 3;
	read_all_key = cmd;
	mod_timer(&cursor_timer, jiffies+interval);
	if (send_key && cmd > KEY_HOME) {
		speakup_fake_key(cmd);
		update_ctr = 3;
	}
}


static int cursor_read_all(struct vc_data *vc, u_short val)
{
	int len, s_index = synth_get_index();
	int index_ctr = last_in, ret = 2;
	if (s_index > 0) {
		while (index_ctr != last_out && markers[index_ctr].index_num != s_index) {
			index_ctr++;
			if (index_ctr >= MARKER_SIZE)
				index_ctr = 0;
		}
		if (index_ctr != last_out) // we found the index
			last_in =index_ctr;
	}
	index_ctr = last_out-last_in;
	if (index_ctr < 0)
		index_ctr += MARKER_SIZE;
	switch (val) {
	case READ_ALL:
		start_read_all_timer(vc, KEY_DOWN, true);
	break;
	case KEY_DOWN:
		if (!cursor_updated && spk_pos == old_cursor_pos)
			ret = 1;
		else if ((len = get_line(vc, spk_y, 0, vc->vc_cols)) <= 0)
			start_read_all_timer(vc, KEY_DOWN, true);
		else {
			say_sentence_buf(currbuf, 0);
			start_read_all_timer(vc, RA_TIMER, false);
		}
		break;
	case RA_TIMER:
		if (index_ctr < 4)
			start_read_all_timer(vc, KEY_DOWN, true);
		else
			start_read_all_timer(vc, RA_TIMER, false);
		break;
	}
	if (ret == 1)
		stop_readall(vc, KT_SPKUP, 0, 0);
	return ret;
}

static int fg_count[16], fg_start[16], fg_end[16];
static int bg_count[16], bg_start[16], bg_end[16];

static void reset_highlight_buffers(struct vc_data *vc)
{
	int i;
	for (i = 0; i < 16; i++) {
		fg_count[i] = 0;
		fg_start[i] = -1;
		fg_end[i] = 0;
		bg_count[i] = 0;
		bg_start[i] = -1;
		bg_end[i] = 0;
	}
}

static int highlight_colors(struct vc_data *vc)
{
	int fg, bg, i, cc = 0;
	u_char attr, ch;

	if (color_count < 2)
		return -1;
	for (i = update_start; i < update_end; i++) {
		ch = get_char(vc, i, &attr);
		if (ch < '0' || ch > 'z') continue;
		bg = (attr & 0xf0) >> 4;
		fg = (attr&0x0f);
		fg_count[fg]++;
		if (fg_start[fg] == -1)
			fg_start[fg] = i;
		fg_end[fg] = i;
		bg_count[bg]++;
		if (bg_start[bg] == -1)
			bg_start[bg] = i;
		bg_end[bg] = i;
	}
	for (i = 0; i < 16; i++) {
		if (fg_count[i] > 0) cc++;
		if (bg_count[i] > 0) cc++;
	}
	return cc;
}

static int fg_scores[] = {
	0, 4, 4, 3,
	4, 3, 3, 1,
	0, 7, 7, 6,
	7, 6, 6, 2
};

static int bg_scores[] = {
	0, 3, 3, 2,
	3, 2, 2, 1,
	0, 5, 5, 4,
	5, 4, 4, 1
};


static int dh_smart(struct vc_data *vc, u_short val)
{
	if (val ==  KEY_LEFT || val ==  KEY_RIGHT) {
		int offset;
		if (val ==  KEY_RIGHT)
			offset = spk_pos - old_cursor_pos;
		else offset = old_cursor_pos - spk_pos;
		if (offset == 1)
			say_char(vc);
		else if (offset < cursor_updated)
			say_from_to(vc, spk_y, 0, vc->vc_cols, 2);
		else say_word(vc);
	} else if (spk_y != old_cursor_y) {
		say_line_from_to(vc, 0, vc->vc_cols, 26);
	} else {
		int i = say_line_from_to(vc, 0, vc->vc_cols, 14);
		if (i == -1)
			say_word(vc);
	}
return 1;
}


static int dh_on(struct vc_data *vc, u_short val)
{
	if (val == KEY_LEFT || val == KEY_RIGHT) {
		say_char(vc);
	} else {
		say_line_from_to(vc, 0, vc->vc_cols, 26);
	}
return 1;
}


static int speak_win(struct vc_data *vc, u_short val)
{
	int line;

	if (win_start < 2) 
		return 0;
	for (line = win_top; line <= win_bottom; line++)
		say_from_to(vc, line, win_left, win_right, 0);
	return 1;
}


static int speak_highlight(struct vc_data *vc, u_short val)
{
	int fg_score, bg_score, fg_n = 0, bg_n = 0;
	int i, fg_pos = 0, bg_pos = 0, start, end;
	int bg_max = 0, fg_max = 0, pos, count = 0;
	u_char attr, attr_match, attr_mask, last_attr = 0;
	char ch, buffer[256];
	if (highlight_colors(vc) < 2)
		return 0;

	for (i = 0; i < 16; i++) {
		if (bg_count[i] > 0) {
			bg_score = bg_scores[i];
			if (bg_score > bg_max) {
				bg_max = bg_score;
				bg_n = bg_count[i];
				bg_pos = i;
			} else if (bg_score == bg_max && bg_count[i] >bg_n) {
				bg_n = bg_count[i];
				bg_pos = i;
			}
		}
		if (fg_count[i] > 0) {
			fg_score = fg_scores[i];
			if (fg_score > fg_max) {
				fg_max = fg_score;
				fg_n = fg_count[i];
				fg_pos = i;
			} else if (fg_score == fg_max && fg_count[i] >fg_n) {
				fg_n = fg_count[i];
			}
		}
	}
	if (fg_max > 2) {
		attr_match = (fg_pos&0x0f);
		attr_mask = 0x0f;
		start = fg_start[fg_pos];
		end = fg_end[fg_pos];
	} else if (bg_max > 1) {
		attr_match = (bg_pos&0x0f)<<4;
		attr_mask = 0xf0;
		start = bg_start[bg_pos];
		end = bg_end[bg_pos];
	} else return 0;
	for (pos = start; pos <= end; pos++) {
		ch = get_char(vc, pos, &attr);
		attr &= attr_mask;
		if (attr == attr_match) {
			if (last_attr != attr_match)
				buffer[count++] = ' ';
			buffer[count++] = ch;
			if (count > 250)
				break;
		}
			last_attr = attr;
	}
	if (count < 2)
		return 0;
	spk_parked |= 0x01;
	spkup_write(buffer, i, spk_punc_mask);
	spk_pos = start;
	spk_x = start%vc->vc_cols;
	spk_y = start/vc->vc_cols;
	return 1;
}

static u_char back_char;

static int say_back_char(struct vc_data *vc, u_short val)
{
	if (KEY_BACKSPACE == val && old_cursor_pos-1 == (vc->vc_pos-vc->vc_origin)/2) {
		speak_char(back_char);
		read_all_key = 0;
	}
	return 1;
}


// must be ordered same as cursor modes in spk_priv_keyinfo.h
static done_hand done_handlers[] = {
	NULL, dh_on, dh_smart, speak_highlight, speak_win,
	NULL, NULL
};

static void cursor_done(u_long data)
{
	struct vc_data *vc;
	unsigned long flags;
	int ret =0;

	del_timer(&cursor_timer);
	if (cursor_con != fg_console)
		return;
	spin_lock_irqsave(&speakup_info.spinlock, flags);
	vc = vc_cons[cursor_con].d;
	current_spk = speakup_console[cursor_con];
	if (read_all_key != READ_MULTI) {
		speakup_date(vc);
		if (cursor_updated || spk_pos != old_cursor_pos)
			update_ctr = 0;
		else if (update_ctr > 0) {
			// nothing happened, wait another timer interval
			update_ctr--;
			start_read_all_timer(vc, read_all_key, false);
			spin_unlock_irqrestore(&speakup_info.spinlock, flags);
			return;
		}
	}
	if (read_all_key >= KEY_HOME && win_enabled) {
		if (vc->vc_x >= win_left && vc->vc_x <= win_right &&
			vc->vc_y >= win_top && vc->vc_y <= win_bottom) {
			goto out;
		}
	}
	if (done_handler)
		ret = (*done_handler)(vc, read_all_key);
	if (!ret)
		dh_on(vc, read_all_key);
out:
	if (ret < 2)
		done_handler = NULL;
	spin_unlock_irqrestore(&speakup_info.spinlock, flags);
}

static int do_cursor(struct vc_data *vc, u_short value)
{
	del_timer(&cursor_timer);
	spk_parked &= 0xf0;
	spk_shut_up &= 0xfe;
	speakup_date(vc);
	update_ctr = 0;
	spk_do_flush();
	if (read_all_mode) {
		last_in = 0;
		last_out = 0;
		if (get_line(vc, spk_y, 0, vc->vc_cols) > 0)
			say_sentence_buf(currbuf, 0);
		done_handler = cursor_read_all;
		start_read_all_timer(vc, READ_ALL, false);
		return (value >= KEY_HOME) ? NOTIFY_OK : NOTIFY_STOP;
	}
	done_handler = NULL;
	if ((spk_shut_up & 0x80) || spk_special_handler ||
		value == KEY_PAGEUP || value == KEY_PAGEDOWN)
		return NOTIFY_OK;
	if (value == KEY_BACKSPACE) {
		if (spk_pos > 0) {
  		back_char = get_char(vc, (spk_pos-1), &spk_attr);
  		done_handler = say_back_char;
		}
	} else if (cursor_track == CT_Off)
		return NOTIFY_OK;
	else {
		done_handler = done_handlers[cursor_track];
		if (win_hdr_len > 0) {
			u_char ch;
			char *cp = win_buf;
			int start = win_header;
			while (*cp) {
				ch = (char) get_char(vc, start++, NULL);
				if (*cp++ != ch) break;
			}
			if (!*cp)
				done_handler = speak_win;
		}
		if (cursor_track == CT_Highlight)
			reset_highlight_buffers(vc);
		else
			get_line(vc, spk_y, 0, vc->vc_cols);
	}
	if (done_handler) {
		start_read_all_timer(vc, value, false);
		update_ctr = 3; // if on a remote connection, wait 3 times for update
	}
	return NOTIFY_OK;
}


/* called by: vt_notifier_call() */
static void speakup_con_write(struct vc_data *vc, const char *str, int len)
{
	unsigned long flags;
	static u_char last_color;

	if (vc->vc_num != fg_console || !synth || speakup_console[fg_console]->shut_up
		|| !spin_trylock_irqsave(&speakup_info.spinlock, flags))
		return;
	current_spk = speakup_console[vc->vc_num];
	if (done_handler) {
		int pos = (vc->vc_pos - vc->vc_origin)/2;
		// this seems to be pre-incremented before we are called
		if (--pos <update_start) {
			if (0x7fff == update_start)// first time in this update
				last_color = vc->vc_attr;
			update_start = pos;
		}
		pos += len;
		if (pos > update_end)
			update_end = pos;
		cursor_updated += len;
		if (last_color != vc->vc_attr) {
			last_color = vc->vc_attr;
			color_count++;
		}
		spin_unlock_irqrestore(&speakup_info.spinlock, flags);
		return;
	}
	if (win_enabled) {
		if (vc->vc_x >= win_left && vc->vc_x <= win_right &&
		    vc->vc_y >= win_top && vc->vc_y <= win_bottom) {
			spin_unlock_irqrestore(&speakup_info.spinlock, flags);
			return;
		}
	}

	spkup_write(str, len, spk_punc_mask);
	spin_unlock_irqrestore(&speakup_info.spinlock, flags);
}

static void speakup_con_update(struct vc_data *vc)
{
	unsigned long flags;
	static int last_console = 0;

	if (speakup_console[vc->vc_num] == NULL ||
		!spin_trylock_irqsave(&speakup_info.spinlock, flags))
		return;
	current_spk = speakup_console[vc->vc_num];
	if (last_console != fg_console) {
		last_console = fg_console;
		console_switch_vars(fg_console);
	}
	if ( !(spk_parked) )
		speakup_date(vc);
	spin_unlock_irqrestore(&speakup_info.spinlock, flags);
}

static void say_status(enum msg_index_t m, int on_off)
{
	char *msg = spk_msg_get(m);
	char *stat = spk_msg_get((on_off) ? MSG_ON : MSG_OFF);
	if (msg&& stat)
		synth_printf("%s, %s\n", msg, stat);
	else if (on_off)
		bleep3(0, -1, 4);
	else bleep3(4, -1, 0);
}

static void do_handle_spec(struct keyboard_notifier_param *param)
{
	unsigned long flags;
	int on_off = 0, vc_num = param->vc->vc_num;
	int value = KVAL(param->value);
	int leds = param->ledstate;
	enum msg_index_t m = 0;
	spin_lock_irqsave(&speakup_info.spinlock, flags);
	current_spk = speakup_console[vc_num];

	switch (value) {
	case KVAL(K_CAPS):
		m = KEY_CAPSLOCK;
		on_off = (leds>>VC_CAPSLOCK)&1;
		break;
	case KVAL(K_NUM):
	case KVAL(K_BARENUMLOCK):
		m = KEY_NUMLOCK;
		on_off = (leds>>VC_NUMLOCK)&1;
		break;
	case KVAL(K_HOLD):
		m = KEY_SCROLLLOCK;
		on_off = (leds>>VC_SCROLLOCK)&1;
		current_spk->tty_stopped = on_off;
		break;
	default:
		spk_parked &= 0xf0;
	}
	if (m) {
		say_status(KEY_START+m, on_off);
	}
	spin_unlock_irqrestore(&speakup_info.spinlock, flags);
}

static void toggle_multi (struct vc_data *vc )
{
	if (spk_adjust_var(0, 0) < 0)
		bleep( 9 );
}

static void set_var (struct vc_data *vc )
{
}

static void route_cursor (struct vc_data *vc )
{
	done_handler = do_route;
	route_x = spk_x;
	route_y = spk_y;
	route_pos = spk_pos;
	spk_parked &= 0xf0;
	start_read_all_timer(vc, READ_ALL, false);
}

static void toggle_column (struct vc_data *vc )
{
	if (current_spk->column_mode >= 0)
		current_spk->column_mode = -1;
	else {
		current_spk->column_mode = spk_y;
	}
	say_status(MSG_COLUMNS, current_spk->column_mode+1);
}

static void toggle_indent (struct vc_data *vc )
{
	current_spk->indent = 1-current_spk->indent;
	say_status(MSG_INDENT, current_spk->indent);
}

static void set_window(struct vc_data *vc)
{
	char info[40];
	char *s_e;

	if (win_start > 1) {
		spk_msg_say(MSG_WINDOW_ALREADY_SET);
		return;
	}
	if (spk_x < win_left || spk_y < win_top) {
		spk_msg_say(MSG_END_BEFORE_START);
		return;
	}
	if (win_start && spk_x == win_left && spk_y == win_top) {
		win_left = 0;
		win_right = vc->vc_cols-1;
		win_bottom = spk_y;
		snprintf(info, sizeof(info), spk_msg_get(MSG_WINDOW_LINE),
			 (int)win_top+1);
	} else {
		if (!win_start) {
			win_top = spk_y;
			win_left = spk_x;
			s_e = spk_msg_get(MSG_START);
		} else {
			win_bottom = spk_y;
			win_right = spk_x;
			s_e = spk_msg_get(MSG_END);
		}
		snprintf(info, sizeof(info), spk_msg_get(MSG_WINDOW_BOUNDARY),
			s_e, (int)spk_y+1, (int)spk_x+1);
	}
	synth_printf("%s\n", info);
	win_start++;
}

static void set_win_header(struct vc_data *vc)
{
	char info[40];
	char *s_e;
	u_char tmp, ch;

	if (win_start < 2) {
		spk_msg_say(MSG_NO_WINDOW);
		return;
	}
	if (win_hdr_len) {
		win_hdr_len = 0;
		win_header = -1;
	}
	if (win_header == -1) {
		win_header = spk_pos;
		s_e = spk_msg_get(MSG_HEADER_START);
	} else {
		char *cp = win_buf;
		int start = win_header;
		int len = spk_pos-start;
		if (len < 1) {
			spk_msg_say(MSG_END_BEFORE_START);
header_cancel:
			win_header = -1;
			win_hdr_len = 0;
			return;
		}
		if (len > HDR_BUF_SIZE-2) {
			spk_msg_say(MSG_WINDOW_LONG_HEADER);
			goto header_cancel;
		}
		if (len == 0)
			goto header_cancel;
		win_hdr_len = len;
		while (--len >= 0) {
			ch = (char) get_char(vc, start++, &tmp);
			*cp++ = ch;
		}
		*cp = '\0';
		s_e = spk_msg_get(MSG_HEADER_END);
	}
	snprintf(info, sizeof(info), spk_msg_get(MSG_WINDOW_BOUNDARY),
			s_e, (int)spk_y+1, (int)spk_x+1);
	synth_printf("%s\n", info);
}

static void clear_window(struct vc_data *vc)
{
	win_top = win_bottom = 0;
	win_left = win_right = 0;
	win_header = -1;
	win_hdr_len = 0;
	win_start = 0;
	spk_msg_say(MSG_WINDOW_CLEARED);
}

static void silence_window(struct vc_data *vc)
{
	if (win_start < 2) {
		spk_msg_say(MSG_NO_WINDOW);
		return;
	}
	win_enabled ^= 1;
	spk_msg_say((win_enabled) ? MSG_WINDOW_SILENCED :
	MSG_WINDOW_SILENCE_DISABLED);
}

static void edit_bits(struct vc_data *vc)
{
	int val = speakup_key-EDIT_SOME;

	if (val < 0 || val > 4) {
		spk_msg_say(MSG_ERROR);
	} else {
	if (val >1)
		val += 2;
	else val++;
		pb_edit = &spk_punc_info[val];
		synth_printf(spk_msg_get(MSG_EDIT_PROMPT), pb_edit->name);
		spk_special_handler = do_edit_bits;
	}
}

static done_hand buffer_handler = NULL;
#define IN_MAX 15
static char in_buff[IN_MAX+1];
static int in_len = 0, in_func = 0, in_allowed = 0;

static int do_getbuff(struct vc_data *vc, u_char type, u_char ch, u_short key)
{
	char *msg;
	if (type == KT_SHIFT || type > 16)
		return 0;
	if (key == KEY_BACKSPACE) {
		if (in_len <= 0)
			return -1;
		ch = in_buff[--in_len];
	} else if (key == KEY_ENTER || ( speakup_key == in_func && in_len == 0)) {
		spk_special_handler = NULL;
		if ( buffer_handler(vc, in_func) >= 0)
		return 1;
		if ((msg = spk_msg_get2(FUNC_ID, in_func)))
			synth_printf(spk_msg_get(MSG_CANCELED), msg);
		return -1;
	} else if (!((1<<type)&in_allowed) || in_len >= IN_MAX) {
		spk_special_handler = NULL;
		if ((msg = spk_msg_get2(FUNC_ID, in_func)))
			synth_printf(spk_msg_get(MSG_CANCELED), msg);
		return -1;
	} else
		in_buff[in_len++] = ch;
	in_buff[in_len] = '\0';
	spkup_write(&ch, 1, spk_punc_mask);
	return 1;
}

static void get_buffer(struct vc_data *vc, int allowed_types, done_hand func)
{
	in_func = speakup_key;
	spk_msg_say2(FUNC_ID, in_func);
	buffer_handler = func;
	in_len = 0;
	*in_buff = '\0';
	in_allowed = allowed_types;
	spk_special_handler = do_getbuff;
}

static int stop_readall(struct vc_data *vc, u_char type, u_char ch, u_short key)
{
	if (read_all_mode != READ_MULTI && key == CURSOR_CHECK)
	return 0;
	del_timer(&cursor_timer);
	if (read_all_mode == READ_MULTI || type != KT_SHIFT) {
		done_handler = NULL;
		spk_special_handler = NULL;
		read_all_mode = 0;
		if (key >0) // no flush from the readall handler
			spk_do_flush();
		return (type == KT_SPKUP) ? 1 : 0;
	}
	read_all_mode ^= PAUSED;
	if ((read_all_mode&PAUSED)) {
		spk_do_flush();
	} else { // start from last index-1
		int n, buf_num = currbuf, pos = 0;
		if (last_in != last_out) {
			n = last_in-1;
			if (n < 0)
				n = MARKER_SIZE-1;
			buf_num = (int)markers[n].buf_num;
			pos = (int)markers[n].pos;
		}
		last_in = 0;
		last_out = 0;
		say_sentence_buf(buf_num, pos);
		while(buf_num != currbuf) {
			buf_num++;
			if (buf_num >= LINEBUFFERS)
				buf_num = 0;
			say_sentence_buf( buf_num, 0);
		}
		start_read_all_timer(vc, READ_ALL, false);
	}
	return 0;
}

static void start_readall(struct vc_data *vc)
{
	if (synth->get_index == NULL) {
		bleep( 9 );
		return;
	}
	read_all_mode =  (speakup_key == READ_ALL_LINE) ?
		LINE_MODE : SENTENCE_MODE;
	spk_special_handler = stop_readall;
	do_cursor(vc, speakup_key);
}

static void set_cursor_mode(struct vc_data *vc)
{
	cursor_track = speakup_key-CURSOR_OFF;
	spk_msg_say2(FUNC_ID, speakup_key);
}

static void speakup_help(struct vc_data *vc)
{
	spk_help_init();
}

static int do_goto(struct vc_data *vc, u_short key)
{
	char *cp;
	u_char go_x = 0, go_y = 0, val;
	if (in_len == 0) {
		go_x = current_spk->goto_x;
		go_y = current_spk->goto_y;
		goto finish_goto;
	}
	for (cp = in_buff; *cp; cp++) {
		cp = spk_s2uchar(cp, &val);
		if (!val)
			return -1;
		if ( 'x' == *cp)
			go_x = val;
		else if ('y' == *cp)
			go_y = val;
		else if ('\0' == *cp && !go_y)
			go_y = val;
		else return -1;
	}
	if (go_x >= vc->vc_cols || go_y >= vc->vc_rows)
		return -1;
	current_spk->goto_x = go_x;
	current_spk->goto_y = go_y;

finish_goto:
	if (go_x > 0)
		spk_x = go_x-1;
	if (go_y > 0)
		spk_y = go_y-1;
	spk_pos = spk_y*vc->vc_cols+spk_x;
	spk_parked |= 0x01;
	if (go_y)
		say_line(vc);
	else say_word(vc);
	return 1;
}

static void speakup_goto(struct vc_data *vc)
{
	get_buffer(vc, 1, do_goto);
}

static void find_up(struct vc_data *vc)
{
	int find_len = current_spk->find_len, tmp_pos = spk_pos;
	char *start  = line_buffer[0]+find_len;
	char *end = start+(2*MAXLINE);
	char *ep = end, *cp = end;
	char ch;
	if (!find_len) {
		spk_msg_say(MSG_NO_BUFF);
		return;
	}
	while (--tmp_pos >= 0) {
		ch = (char) get_char(vc, tmp_pos, NULL);
		*--cp = ch;
		if (ep-cp <= find_len) continue;
		if (!memcmp(cp, current_spk->find_buff, find_len))
			break;
		ep--;
		if (cp <= start) { // move start of buffer up
			char *bp = end;
			while (ep >= cp) *bp-- = *ep--;
			ep = end;
			cp = bp+1;
		}
	}
	if (tmp_pos >= 0) {
		spk_parked |= 0x01;
		spk_pos = tmp_pos;
		spk_x = tmp_pos%vc->vc_cols;
		spk_y = tmp_pos/vc->vc_cols;
		say_word(vc);
	} else 
		spk_msg_say(MSG_NOT_FOUND);
}

static void find_down(struct vc_data *vc)
{
	int find_len = current_spk->find_len, tmp_pos = spk_pos;
	int screen_end = (vc->vc_screenbuf_size/2)-find_len;
	char *start  = line_buffer[0];
	char *end = start+(2*MAXLINE);
	char *fp = start, *cp = start;
	char ch;
	if (!find_len) {
		spk_msg_say(MSG_NO_BUFF);
		return;
	}
	while (++tmp_pos < screen_end) {
		ch = (char) get_char(vc, tmp_pos, NULL);
		*cp++ = ch;
		if (cp-fp <= find_len) continue;
		if (!memcmp(fp, current_spk->find_buff, find_len))
			break;
		fp++;
		if (cp > end) { // move end of buffer down
			char *bp = start;
			while (fp < cp) *bp++ = *fp++;
			fp = start;
			cp = bp;
		}
	}
	if (tmp_pos < screen_end) {
		spk_parked |= 0x01;
		tmp_pos -= find_len;
		spk_pos = tmp_pos;
		spk_x = tmp_pos%vc->vc_cols;
		spk_y = tmp_pos/vc->vc_cols;
		say_word(vc);
	} else 
		spk_msg_say(MSG_NOT_FOUND);
}

static int do_find(struct vc_data *vc, u_short key)
{
	if (in_len <= 0)
		return -1;
	memcpy( current_spk->find_buff, in_buff, in_len+1); // with the null byte
	current_spk->find_len = in_len;
	if (key == FIND_PREV)
		find_up(vc);
	else find_down(vc);
	return 1;
}

static void speakup_find(struct vc_data *vc)
{
	get_buffer(vc, 1, do_find);
}

static void speakup_lock(struct vc_data *vc)
{
					spk_key_locked ^= 16;
	key_speakup = spk_key_locked;
	bleep((spk_key_locked) ? 12 : 0);
}

spkup_hand spkup_handlers[FIRST_ADJUST_VAR] = {
	[SPEECH_KILL] = speech_kill,
	[SPEAKUP_QUIET] = speakup_shut_up,
	[SPK_LOCK] = speakup_lock,
	[SPEAKUP_HELP] = speakup_help,
	[SAY_CHAR_ABOVE] = say_char_above,
	[SAY_CHAR_BELOW] = say_char_below,
	[SAY_FIRST_CHAR] = say_first_char,
	[SAY_LAST_CHAR] = say_last_char,
	[SAY_CHAR] = say_char,
	[SAY_PREV_CHAR] = say_prev_char,
	[SAY_NEXT_CHAR] = say_next_char,
	[SAY_WORD] = say_word,
	[SAY_PREV_WORD] = say_prev_word,
	[SAY_NEXT_WORD] = say_next_word,
	[SAY_WORD_ABOVE] = say_word_above,
	[SAY_WORD_BELOW] = say_word_below,
	[SAY_LINE] = say_line,
	[SAY_PREV_LINE] = say_prev_line,
	[SAY_NEXT_LINE] = say_next_line,
	[SAY_LINE_NC] = say_line,
	[SAY_PREV_LINE_NC] = say_prev_line,
	[SAY_NEXT_LINE_NC] = say_next_line,
	[TOP_EDGE] = top_edge,
	[BOTTOM_EDGE] = bottom_edge,
	[LEFT_EDGE] = left_edge,
	[RIGHT_EDGE] = right_edge,
	[SPELL_PHONETIC] = spell_word,
	[SPELL_WORD] = spell_word,
	[WORD_END] = find_word_end,
	[SAY_SCREEN] = say_screen,
	[SAY_POSITION] = say_position,
	[SAY_LINE_COLUMN] = say_position,
	[SAY_ATTRIBUTES] = say_attributes,
	[SPEAKUP_OFF] = speakup_off,
	[SPEAKUP_PARKED] = speakup_parked,
	[SAY_LINE_INDENT] = say_line,
	[SAY_FROM_TOP] = say_from_top,
	[SAY_TO_BOTTOM] = say_to_bottom,
	[SAY_FROM_LEFT] = say_from_left,
	[SAY_TO_RIGHT] = say_to_right,
	[SAY_CHAR_NUM] = say_char_num,
	[SAY_PHONETIC_CHAR] = say_phonetic_char,
	[EDIT_SOME] = edit_bits,
	[EDIT_MOST] = edit_bits,
	[EDIT_DELIM] = edit_bits,
	[EDIT_REPEAT] = edit_bits,
	[EDIT_EXNUM] = edit_bits,
	[SET_WIN] = set_window,
	[CLEAR_WIN] = clear_window,
	[SET_WIN_HEADER] = set_win_header,
	[ENABLE_WIN] = silence_window,
	[SAY_WIN] = say_window,
	[SET_VAR] = set_var,
	[ROUTE_CURSOR] = route_cursor,
	[TOGGLE_COLUMN] = toggle_column,
	[TOGGLE_CURSORING] = toggle_cursoring,
	[TOGGLE_INDENT] = toggle_indent,
	[TOGGLE_MULTI] = toggle_multi,
	[CURSOR_OFF] = set_cursor_mode,
	[CURSOR_ON] = set_cursor_mode,
	[CURSOR_SMART] = set_cursor_mode,
	[CURSOR_HIGHLIGHT] = set_cursor_mode,
	[CURSOR_WIN] = set_cursor_mode,
	[READ_ALL] = start_readall,
	[READ_ALL_LINE] = start_readall,
	[CUT] = speakup_cut,
	[CUT1] = speakup_cut,
	[CUT2] = speakup_cut,
	[CUT3] = speakup_cut,
	[PASTE] = speakup_paste,
	[PASTE1] = speakup_paste,
	[PASTE2] = speakup_paste,
	[PASTE3] = speakup_paste,
	[PASTE_READ] = read_paste,
	[PASTE_READ1] = read_paste,
	[PASTE_READ2] = read_paste,
	[PASTE_READ3] = read_paste,
	[SET_VARS] = spk_set_vars,
	[GOTO] = speakup_goto,
	[FIND_PREV] = speakup_find,
	[FIND_NEXT] = speakup_find,
	[FIND_UP] = find_up,
	[FIND_DOWN] = find_down
};

static int do_spkup(struct vc_data *vc, u_char value)
{
	spk_lastkey = 0;
	if (value < FIRST_ADJUST_VAR) {
		// save pos so if we move virtual cursor, we calculate x and y
		int prev_spk_pos = spk_pos;
		if (done_handler) { // kill pending cursor_done
			del_timer(&cursor_timer);
			done_handler = NULL;
		}
		spk_do_flush();
	  if (spkup_handlers[value])
			(*spkup_handlers[value]) (vc);
		else synth_printf("handler %d not implemented yet\n", (int)value);
		if (spk_pos != prev_spk_pos) {
			spk_x = spk_pos%vc->vc_cols;
			spk_y = spk_pos/vc->vc_cols;
			spk_parked |= 0x01;
		}
	} else if (value < SPKUP_MAX_KEY) {
		int var_id = value - FIRST_ADJUST_VAR;
		int inc = (var_id&1) ? 1 : -1;
		if (spk_adjust_var((var_id/2)+FIRST_SET_VAR, inc) < 0)
			bleep(9);
	} else {
		synth_printf("do_speakup, bad value %d\n", (int)value );
			bleep(13);
	}
	return NOTIFY_STOP;
}

static int keycode;

static int
handle_keycode(struct keyboard_notifier_param *param)
{
	static u_short last_keycode = 0;
	static u_long last_jiffy = 0;
	unsigned long flags;
	struct vc_data *vc = param->vc;
	u_char *key_info;
	u_char new_key = 0, shift_info, offset;
	int ret = NOTIFY_OK;
	speakup_key = 0;

	// check as much as we can without locking
	keycode = param->value;
	if (param->down) {
		spk_close_press = (last_keycode == keycode &&
		    time_after(last_jiffy + MAX_DELAY, jiffies)) ? 1 : 0;
		last_keycode = keycode;
		last_jiffy = jiffies;
	} 
	if (keycode >= MAX_KEY || !(key_info = spk_our_keys[keycode]) )
		return ret;
	shift_info = (param->shift&0x0f) + key_speakup;
	offset = spk_shift_table[shift_info];
	if (!offset || !(new_key = key_info[offset]) )
		return ret;
	if (new_key != SPEECH_KILL && (spk_killed))
		return ret;
	if (!param->down) {
		if 	(SPK_KEY == new_key && !spk_key_locked)
			key_speakup = 0;
		return ret;
	}
	spin_lock_irqsave(&speakup_info.spinlock, flags);
	current_spk = speakup_console[vc->vc_num];
	if 	(SPK_KEY == new_key) {
		key_speakup = 16;
		if ( !spk_special_handler) {
			ret = NOTIFY_STOP;
			spk_shut_up &= 0xfe;
			spk_do_flush();
		}
		speakup_key = new_key;
		goto out;
	}
	if (CURSOR_CHECK == new_key) {
		ret = do_cursor(vc, keycode);
		goto out;
	}
	if (spk_close_press) {
		offset = spk_shift_table[shift_info + 32];
		// double press?
		if (offset && key_info[offset])
			new_key = key_info[offset];
	}
	speakup_key = new_key;
	spk_shut_up &= 0xfe;
	if ( !spk_special_handler)
		ret = do_spkup(vc, new_key);
out:
	spin_unlock_irqrestore(&speakup_info.spinlock, flags);
	return ret;
}

#define NO_FLUSH 0xe4 // types spec, cons, cur and shift
// types which do not flush synth if no special handler

static int
handle_keysym(struct keyboard_notifier_param *param)
{
	static const char *pad_chars = "0123456789+-*/\015,.?()";
	struct vc_data *vc = param->vc;
	unsigned long flags;
	int ret = 0, keysym = param->value, check;
	u_char type, value;
	if (speakup_key) {
		type = KT_SPKUP;
		value = speakup_key;
	} else {
		type = KTYP(keysym)-0xf0;
		value = KVAL(keysym);
	}
	if (type == KT_SPEC && value == 1) {
		value = '\n';
		type = KT_LATIN;
	} else if (type == KT_PAD && ((param->ledstate>>VC_NUMLOCK)&1)) {
		value = spk_lastkey = pad_chars[value];
		type = KT_LATIN;
	} else if (type == KT_LETTER) {
		type = KT_LATIN;
	}
	if (!param->down) {
		if (type == KT_LATIN)
			spk_lastkey = 0;
		return NOTIFY_OK;
	}
	check = 1<<type;
	if (!spk_special_handler && (check&NO_FLUSH))
		return NOTIFY_OK;

	spin_lock_irqsave(&speakup_info.spinlock, flags);
	current_spk = speakup_console[vc->vc_num];
	if (!spk_special_handler) {
		spk_shut_up &= 0xfe;
		spk_parked &= 0xf0;
		if (!spk_no_intr)
			spk_do_flush();
		if (type == KT_LATIN) {
			spk_lastkey = value;
			if (spk_key_echo == 2 && value >= MINECHOCHAR)
				speak_char(value);
			if (spk_bell_pos && (vc->vc_x == spk_bell_pos-1))
				bleep(3);
		}
		spin_unlock_irqrestore(&speakup_info.spinlock, flags);
		return NOTIFY_OK;
	}
	if (value == 0x7f)
		value = 8;	// make del = backspace
	if (param->down > 1 && (KT_SHIFT == type || KT_SPKUP == type))
		ret = 1; // kill repeats
	else {
		spk_shut_up &= 0xfe;
		spk_do_flush();
		ret = (*spk_special_handler) (vc, type, value, keycode);
	}
	if (ret < 0)
		bleep(9);
	spin_unlock_irqrestore(&speakup_info.spinlock, flags);
	return (!ret) ? NOTIFY_OK : NOTIFY_STOP;
}

static int keyboard_notifier_call(struct notifier_block *nb,
				  unsigned long code, void *_param)
{
	struct keyboard_notifier_param *param = _param;
	struct vc_data *vc = param->vc;
	int ret = NOTIFY_OK;

	if (vc->vc_mode == KD_GRAPHICS)
		return ret;

// if we have a fake key press from readall, let the system handle it.
	if (speakup_fake_key_pressed() || synth == NULL)
		return ret;

	switch (code) {
	case KBD_KEYCODE:
		ret = handle_keycode(param);
		break;
	case KBD_UNBOUND_KEYCODE:
		break;
	case KBD_UNICODE:
		break;
	case KBD_KEYSYM:
		if (!(spk_killed))
			ret = handle_keysym(param);
		break;
	case KBD_POST_KEYSYM:
		if (param->down && !(spk_killed)) {
			unsigned char type = KTYP(param->value)-0xf0;
			switch (type) {
				case KT_SHIFT:
					do_handle_shift(param);
					break;
				case KT_SPEC:
					do_handle_spec(param);
			}
		}
		break;
	}
	return ret;
}

static int vt_notifier_call(struct notifier_block *nb,
			    unsigned long code, void *_param)
{
	struct vt_notifier_param *param = _param;
	struct vc_data *vc = param->vc;

	switch (code) {
	case VT_ALLOCATE:
		if (vc->vc_mode == KD_TEXT)
			speakup_allocate(vc);
		break;
	case VT_DEALLOCATE:
		speakup_deallocate(vc);
		break;
	case VT_WRITE:
		if (param->c < 0x100) {
			char d = param->c;
			if ( '\b' == d )
				say_back_char(vc, read_all_key);
			else {
				if ('\t' == d) d = ' ';
				speakup_con_write(vc, &d, 1);
			}
		}
		break;
	case VT_UPDATE:
		speakup_con_update(vc);
		break;
	}
	return NOTIFY_OK;
}

/* called by: module_exit() */
static void __exit speakup_exit(void)
{
	int i;

	unregister_keyboard_notifier(&keyboard_notifier_block);
	unregister_vt_notifier(&vt_notifier_block);
	speakup_unregister_devsynth();
	del_timer(&cursor_timer);
	kthread_stop(speakup_task);
	speakup_task = NULL;
	mutex_lock(&spk_mutex);
	synth_release();
	mutex_unlock(&spk_mutex);

	speakup_kobj_exit();

	for (i = 0; i < MAX_NR_CONSOLES; i++)
		kfree(speakup_console[i]);

	speakup_remove_virtual_keyboard();
	speakup_free_selection();

	for (i = 0; i < MAXVARS; i++)
		speakup_unregister_var(i);

	for (i = 0; i < 256; i++) {
		if (spk_characters[i] != spk_default_chars[i])
			kfree(spk_characters[i]);
	}

	spk_free_user_msgs();
}

/* call by: module_init() */
static int __init speakup_init(void)
{
	int i;
	long err = 0;
	struct vc_data *vc = vc_cons[fg_console].d;

	/* These first few initializations cannot fail. */
	speakup_init_vars(vc->vc_cols);
	spk_initialize_msgs();	/* Initialize arrays for i18n. */
	spk_reset_default_chars();
	spk_reset_default_chartab();
	spk_strlwr(synth_name);
	for (i = 1; spk_punc_info[i].mask != 0; i++)
		spk_set_mask_bits(NULL, i, 2);

	spk_set_key_info(spk_key_defaults);
	speakup_init_selection( );

	/* From here on out, initializations can fail. */
	err = speakup_add_virtual_keyboard();
	if (err)
		goto error_virtkeyboard;

	current_spk = kzalloc(sizeof(*current_spk), GFP_KERNEL);
	if (!current_spk) {
		err = -ENOMEM;
		goto error_alloc;
	}

	cursor_track = 1;
	current_spk->column_mode = -1;
	speakup_console[vc->vc_num] = current_spk;
	speakup_date(vc);

	for (i = 0; i < MAX_NR_CONSOLES; i++)
		if (vc_cons[i].d) {
			err = speakup_allocate(vc_cons[i].d);
			if (err)
				goto error_kobjects;
		}

	if (spk_quiet_boot)
		spk_shut_up |= 0x01;

	err = speakup_kobj_init();
	if (err)
		goto error_kobjects;

	synth_init(synth_name);
	speakup_register_devsynth();
	/*
	 * register_devsynth might fail, but this error is not fatal.
	 * /dev/synth is an extra feature; the rest of Speakup
	 * will work fine without it.
	 */

	err = register_keyboard_notifier(&keyboard_notifier_block);
	if (err)
		goto error_kbdnotifier;
	err = register_vt_notifier(&vt_notifier_block);
	if (err)
		goto error_vtnotifier;

	speakup_task = kthread_create(speakup_thread, NULL, "speakup");

	if (IS_ERR(speakup_task)) {
		err = PTR_ERR(speakup_task);
		goto error_task;
	}

	set_user_nice(speakup_task, 10);
	wake_up_process(speakup_task);

	pr_info("speakup %s: initialized\n", SPEAKUP_VERSION);
	pr_info("synth name on entry is: %s\n", synth_name);
	goto out;

error_task:
	unregister_vt_notifier(&vt_notifier_block);

error_vtnotifier:
	unregister_keyboard_notifier(&keyboard_notifier_block);
	del_timer(&cursor_timer);

error_kbdnotifier:
	speakup_unregister_devsynth();
	mutex_lock(&spk_mutex);
	synth_release();
	mutex_unlock(&spk_mutex);
	speakup_kobj_exit();

error_kobjects:
	for (i = 0; i < MAX_NR_CONSOLES; i++)
		kfree(speakup_console[i]);

error_alloc:
	speakup_remove_virtual_keyboard();

error_virtkeyboard:
	for (i = 0; i < MAXVARS; i++)
		speakup_unregister_var(i);

	for (i = 0; i < 256; i++) {
		if (spk_characters[i] != spk_default_chars[i])
			kfree(spk_characters[i]);
	}

	spk_free_user_msgs();

out:
	return err;
}

module_init(speakup_init);
module_exit(speakup_exit);
