/*
 * originally written by: Kirk Reiser <kirk@braille.uwo.ca>
 * this version considerably modified by David Borowski, david.a.borowski@gmail.com
 *
 * Copyright (C) 1998-99  Kirk Reiser.
 * Copyright (C) 2003 David Borowski.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * specificly written as a driver for the speakup screenreview
 * s not a general device driver.
 */
#include <linux/unistd.h>
#include <linux/proc_fs.h>
#include <linux/jiffies.h>
#include <linux/spinlock.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/kthread.h>
#include "speakup.h"
#include "spk_priv.h"
#include "serialio.h"

#define DRV_VERSION "2.20"
#define SYNTH_CLEAR 0x03
#define PROCSPEECH 0x0b
static int xoff;

static void do_catch_up(struct spk_synth *synth);
static void synth_flush(struct spk_synth *synth);
static void read_buff_add(u_char c);
static unsigned char get_index(void);
static int new_voice(int val);

static int is_flushing = 0;
static spinlock_t flush_lock;

static struct var_t vars[] = {
	{ CAPS_START, .u.s = {"[:dv ap 140] " } },
	{ CAPS_STOP, .u.s = {"[:dv ap 100 ] " } },
	{ RATE, .u.n = {"[:ra %d]", 32, 0, 50, 300, 5, NULL }},
	{ VOICE, .u.n = {"[:n%c] ", 0, 0, 9, 0, 0, "phfdburwkv" } },
	{ PITCH, .u.n = {"[:dv ap %d]", 12, 0, 54, 40, 5, NULL }},
	{ VOL, .u.n = {"[:dv gv %d]", 12, 0, 15, 40, 2, NULL }},
	{ PUNCT, .u.n = {"[:pu %c] ", 0, 0, 2, 0, 0, "nsa" } },
	{ DIRECT, .u.n = {NULL, 0, 0, 1, 0, 0, NULL } },
	V_LAST_VAR
};

/*
 * These attributes will appear in /sys/accessibility/speakup/dectlk.
 */
static struct kobj_attribute caps_start_attribute =
	__ATTR(caps_start, S_IWUSR|S_IRUGO, spk_var_show, spk_var_store);
static struct kobj_attribute caps_stop_attribute =
	__ATTR(caps_stop, S_IWUSR|S_IRUGO, spk_var_show, spk_var_store);
static struct kobj_attribute pitch_attribute =
	__ATTR(pitch, S_IWUSR|S_IRUGO, spk_var_show, spk_var_store);
static struct kobj_attribute punct_attribute =
	__ATTR(punct, S_IWUSR|S_IRUGO, spk_var_show, spk_var_store);
static struct kobj_attribute rate_attribute =
	__ATTR(rate, S_IWUSR|S_IRUGO, spk_var_show, spk_var_store);
static struct kobj_attribute voice_attribute =
	__ATTR(voice, S_IWUSR|S_IRUGO, spk_var_show, spk_var_store);
static struct kobj_attribute vol_attribute =
	__ATTR(vol, S_IWUSR|S_IRUGO, spk_var_show, spk_var_store);

static struct kobj_attribute delay_time_attribute =
	__ATTR(delay_time, S_IWUSR|S_IRUGO, spk_var_show, spk_var_store);
static struct kobj_attribute direct_attribute =
	__ATTR(direct, S_IWUSR|S_IRUGO, spk_var_show, spk_var_store);
static struct kobj_attribute full_time_attribute =
	__ATTR(full_time, S_IWUSR|S_IRUGO, spk_var_show, spk_var_store);
static struct kobj_attribute jiffy_delta_attribute =
	__ATTR(jiffy_delta, S_IWUSR|S_IRUGO, spk_var_show, spk_var_store);
static struct kobj_attribute trigger_time_attribute =
	__ATTR(trigger_time, S_IWUSR|S_IRUGO, spk_var_show, spk_var_store);

/*
 * Create a group of attributes so that we can create and destroy them all
 * at once.
 */
static struct attribute *synth_attrs[] = {
	&caps_start_attribute.attr,
	&caps_stop_attribute.attr,
	&pitch_attribute.attr,
	&punct_attribute.attr,
	&rate_attribute.attr,
	&voice_attribute.attr,
	&vol_attribute.attr,
	&delay_time_attribute.attr,
	&direct_attribute.attr,
	&full_time_attribute.attr,
	&jiffy_delta_attribute.attr,
	&trigger_time_attribute.attr,
	NULL,	/* need to NULL terminate the list of attributes */
};

static int pitch_defaults[] = { 14, 10, 23, 14, 34, 40, 32, 13, 53, 53};
static int vol_defaults[] = { 13, 12, 13, 13, 12, 12, 13, 13, 11, 13};

static struct spk_synth synth_dectlk = {
	.name = "dectlk",
	.version = DRV_VERSION,
	.long_name = "Dectalk Express",
	.init = "[:dv ap 100][:error sp][:pe -380][:po sl 300]",
	.procspeech = PROCSPEECH,
	.clear = SYNTH_CLEAR,
	.delay = 500,
	.trigger = 50,
	.jiffies = 50,
	.full = 5000,
	.flags = SF_DEC,
	.startup = SYNTH_START,
	.checkval = SYNTH_CHECK,
	.vars = vars,
	.probe = spk_serial_synth_probe,
	.release = spk_serial_release,
	.synth_immediate = spk_synth_immediate,
	.catch_up = do_catch_up,
	.flush = synth_flush,
	.is_alive = spk_synth_is_alive_restart,
	.new_voice = new_voice,
	.read_buff_add = read_buff_add,
	.get_index = get_index,
	.indexing = {
		.command = "[:in re %d ] ",
		.lowindex = 1,
		.highindex = 99,
		.currindex = 1,
	},
	.attributes = {
		.attrs = synth_attrs,
		.name = "dectlk",
	},
};

static u_char lastind;

static unsigned char get_index(void)
{
	u_char rv;

	rv = lastind;
	lastind = 0;
	return rv;
}

static void read_buff_add(u_char c)
{
	static int ind = -1;

	if (c == 0x01) {
		unsigned long flags;
		ind = -1;
		xoff = 0;

		spin_lock_irqsave(&flush_lock, flags);
		is_flushing = 0;
		spin_unlock_irqrestore(&flush_lock, flags);
	} else if (c == 0x13) {
		xoff = 1;
	} else if (c == 0x11) {
		xoff = 0;
	} else if (c >= '0' && c <= '9') {
		c &= 15;
		if (ind == -1)
			ind = c;
		else
			ind = ind * 10 + c;
	} else if ((c > 31) && (c < 127)) {
		if (ind != -1)
			lastind = (u_char)ind;
		ind = -1;
	}
}


static void send_command(struct spk_synth *synth, unsigned long flags)
{ // get whole command from buffer, don't allow flush
	int i = 0, n = 0;
	char buffer[16], ch;
	buffer[n++] = '[';
	do {
		ch = synth_buffer_getc();
		buffer[n++] = ch;
	} while (n < 15 && ch != ']');
	spin_unlock_irqrestore(&speakup_info.spinlock, flags);
	if (ch != ']')
		return; // oops, no command end
	while (i < n) {
		if (xoff || !spk_serial_out(buffer[i])) {
			set_current_state(TASK_UNINTERRUPTIBLE);
			schedule_timeout(spk_jiffy_time);
			set_current_state(TASK_RUNNING);
			if (!synth->alive)
				break;
			continue;
		}
		i++;
	}
}

static void do_catch_up(struct spk_synth *synth)
{
	static u_char last = '\0';
	u_char ch;
	unsigned long flags, jiff_max = jiffies + spk_jiffy_time;
	int status;

	while (!kthread_should_stop()) {
		spin_lock_irqsave(&speakup_info.spinlock, flags);
		if ((status = synth_buffer_status(true)) <= 0) {
			spin_unlock_irqrestore(&speakup_info.spinlock, flags);
			if (status == 0)
				break;
			synth->flush(synth);
			continue;
		}
		ch = synth_buffer_getc();
		if (ch == '[') {
			send_command(synth, flags);
			continue;
		}
		spin_unlock_irqrestore(&speakup_info.spinlock, flags);
		set_current_state(TASK_INTERRUPTIBLE);
		if (ch == '\n')
			ch = 0x0D;
		if (xoff || !spk_serial_out(ch)) {
			spin_lock_irqsave(&speakup_info.spinlock, flags);
			synth_buffer_ungetc();
			spin_unlock_irqrestore(&speakup_info.spinlock, flags);
			schedule_timeout(spk_full_time);
			continue;
		}
		set_current_state(TASK_RUNNING);
		if (ch <= SPACE) {
			if (time_after_eq(jiffies, jiff_max)) {
				schedule_timeout(spk_delay_time);
				jiff_max = jiffies + spk_jiffy_time;
			} else if (spk_chartab[last] == A_PUNC)
				spk_serial_out(PROCSPEECH);
		}
		last = ch;
	}
	spk_serial_out(PROCSPEECH);
}

static void synth_flush(struct spk_synth *synth)
{
	unsigned long flags;
	int n, ctr = 4;
	spin_lock_irqsave(&flush_lock, flags);
		is_flushing = 1;
	spin_unlock_irqrestore(&flush_lock, flags);
	spk_serial_out(SYNTH_CLEAR);
	do {
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(spk_jiffy_time/2);
		set_current_state(TASK_RUNNING);
		spin_lock_irqsave(&flush_lock, flags);
		n = is_flushing;
		spin_unlock_irqrestore(&flush_lock, flags);
	} while (n && --ctr > 0);
}

static int new_voice(int val)
{
 char buf[32];
	int len, pitch = pitch_defaults[val];
	struct st_var_header *p_header;
	struct var_t *var_data;
	p_header = spk_get_var_header(PITCH);
	if (!p_header) return -1;
	spk_set_num_var(pitch, p_header, E_NEW_DEFAULT );
	var_data = p_header->data;
		if (var_data->u.n.multiplier)
			pitch *= var_data->u.n.multiplier;
	pitch += var_data->u.n.offset;
	len = sprintf( buf, "[:dv ap %d]", pitch );
	p_header = spk_get_var_header(CAPS_STOP);
	if (!p_header) return -1;
	spk_set_string_var(buf, p_header, len );
	len = sprintf( buf, "[:dv ap %d]", pitch+50 );
	p_header = spk_get_var_header(CAPS_START);
	if (!p_header) return -1;
	spk_set_string_var(buf, p_header, len );
	p_header = spk_get_var_header(VOL);
	if (!p_header) return -1;
	spk_set_num_var(vol_defaults[val], p_header, E_NEW_DEFAULT );
	return 0;
}

module_param_named(ser, synth_dectlk.ser, int, S_IRUGO);
module_param_named(start, synth_dectlk.startup, short, S_IRUGO);

MODULE_PARM_DESC(ser, "Set the serial port for the synthesizer (0-based).");
MODULE_PARM_DESC(start, "Start the synthesizer once it is loaded.");

module_spk_synth(synth_dectlk);

MODULE_AUTHOR("Kirk Reiser <kirk@braille.uwo.ca>");
MODULE_AUTHOR("David Borowski");
MODULE_DESCRIPTION("Speakup support for DECtalk Express synthesizers");
MODULE_LICENSE("GPL");
MODULE_VERSION(DRV_VERSION);

